"""
    domonic.CDN
    ====================================
    For quick reference when prototyping
"""

class CDN_JS:
    """
    JavaScript libraries
    """

    JQUERY: str = "https://code.jquery.com/jquery-3.6.4.min.js"
    JQUERY_UI: str = "https://code.jquery.com/ui/1.13.2/jquery-ui.min.js"
    UNDERSCORE: str = "https://cdn.jsdelivr.net/npm/underscore@1.13.6/underscore-min.js"
    BOOTSTRAP: str = "https://stackpath.bootstrapcdn.com/bootstrap/5.3.0/js/bootstrap.min.js"
    POPPER: str = "https://cdn.jsdelivr.net/npm/popper.js@2.11.7/dist/umd/popper.min.js"
    D3: str = "https://cdnjs.cloudflare.com/ajax/libs/d3/7.8.4/d3.min.js"
    MODERNIZER: str = "https://cdnjs.cloudflare.com/ajax/libs/modernizr/3.11.7/modernizr.min.js"
    MOMENT: str = "https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.29.4/moment.min.js"
    PIXI: str = "https://cdnjs.cloudflare.com/ajax/libs/pixi.js/7.1.0/pixi.min.js"
    SOCKET: str = "https://cdnjs.cloudflare.com/ajax/libs/socket.io/4.6.1/socket.io.min.js"
    X3DOM: str = "https://www.x3dom.org/download/x3dom.js"
    AFRAME: str = "https://aframe.io/releases/1.3.0/aframe.min.js"
    BRYTHON: str = "https://cdnjs.cloudflare.com/ajax/libs/brython/3.11.1/brython.min.js"
    MATHML: str = "https://cdnjs.cloudflare.com/ajax/libs/mathjax/3.2.2/es5/tex-mml-chtml.min.js"
    HTMX: str = "https://unpkg.com/htmx.org@1.9.0"
    LODASH: str = "https://cdn.jsdelivr.net/npm/lodash@4.17.21/lodash.min.js"
    AXIOS: str = "https://cdn.jsdelivr.net/npm/axios@0.21.1/dist/axios.min.js"
    DAY_JS: str = "https://cdn.jsdelivr.net/npm/dayjs@1.10.4/dayjs.min.js"
    CHART_JS: str = "https://cdn.jsdelivr.net/npm/chart.js@2.9.4/dist/Chart.min.js"
    ANIME_JS: str = "https://cdn.jsdelivr.net/npm/animejs@3.2.1/lib/anime.min.js"
    VALIDATOR_JS: str = "https://cdn.jsdelivr.net/npm/validator@13.6.0/validator.min.js"

class CDN_CSS:
    """
    CSS Libraries
    """

    BOOTSTRAP: str = "https://stackpath.bootstrapcdn.com/bootstrap/5.3.0/css/bootstrap.min.css"
    MARX: str = "https://unpkg.com/marx-css/css/marx.min.css"
    MVP: str = "https://unpkg.com/mvp.css"
    WATER: str = "https://cdn.jsdelivr.net/gh/kognise/water.css@latest/water.min.css"
    BALLOON: str = "https://unpkg.com/balloon-css/balloon.min.css"
    THREE_DOTS: str = "https://cdnjs.cloudflare.com/ajax/libs/three-dots/0.2.0/three-dots.min.css"
    MILLIGRAM: str = "https://cdnjs.cloudflare.com/ajax/libs/milligram/1.3.0/milligram.css"
    X3DOM: str = "https://www.x3dom.org/download/x3dom.css"
    FONTAWESOME: str = "https://use.fontawesome.com/releases/v5.7.1/css/all.css"
    MDI: str = "https://cdn.materialdesignicons.com/5.4.55/css/materialdesignicons.min.css"
    TAILWIND: str = "https://unpkg.com/tailwindcss@^2.2.15/dist/tailwind.min.css"
    SIMPLE: str = "https://cdn.simplecss.org/simple.min.css"


class CDN_IMG:
    """CDN images"""

    # - icons
    # - UI - emojis

    """
    # SOME EXAMPLES. NOT ALL ARE HTTPS:
    http://placehold.it/350x150
    http://unsplash.it/200/300
    http://lorempixel.com/400/200
    http://dummyimage.com/600x300/000/fff
    # https://dummyimage.com/420x320/ff7f7f/333333.png&text=Sample
    http://placekitten.com/200/300
    https://placeimg.com/640/480/any
    http://placebear.com/g/200/300
    https://ipsumimage.appspot.com/140x100, ff7700
    https://www.fillmurray.com/640/360
    https://baconmockup.com/640/360
    https://placebeard.it/640x360
    https://www.placecage.com/640/360
    https://www.stevensegallery.com/640/360
    https://fakeimg.pl/640x360
    # https://fakeimg.pl/420x320/ff0000,128/333333,255/?text=Sample&font=lobster
    https://picsum.photos/640/360
    https://via.placeholder.com/420x320/ff7f7f/333333?text=Sample
    https://keywordimg.com/420x320/random
    http://www.dummysrc.com/430x320.png/22c5fc/17202A
    """
    PLACEHOLDER_SERVICE: str = "loremflickr.com"

    @staticmethod
    def PLACEHOLDER(width: int = 100, height: int = 100, HTTP: str = "", seperator: str = "/") -> str:
        """
        to update do CDN_IMG.PLACEHOLDER_SERVICE = "placebear.com/g"
        usage : img(_src=CDN_IMG.PLACEHOLDER(300,100))
        default HTTP is none, to let the browser decide
        # use optional seperator if the site uses x instead of slash
        img(_src=CDN_IMG.PLACEHOLDER(300,100,'x'))
        """
        return f"{HTTP}://{CDN_IMG.PLACEHOLDER_SERVICE}/{width}{seperator}{height}"


class CDN_FONT:
    @staticmethod
    def google(family: str) -> str:
        """pass a font family name and returns the url"""
        return "https://fonts.googleapis.com/css?family=" + "+".join(family)

    # @staticmethod
    # def font_awesome(version='5.7.1'):
    #     return f"https://use.fontawesome.com/releases/v{version}/css/all.css"


# class CDN_TEXT:
# lorem ipusm generator
# fake names, addresses

# class CDN_VIDEO:
# class CDN_AUDIO:



"""
    domonic.decorators
    ====================================

"""

import functools
import inspect
import io
import logging
import warnings
from typing import Callable


def _invoke_before(before, function):
    result = before()
    params = list(inspect.signature(function).parameters.values())
    can_receive_value = any(
        param.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
        for param in params
    ) or any(param.kind == inspect.Parameter.VAR_POSITIONAL for param in params)
    return result, can_receive_value

def el(element="div", string: bool = False):
    """[wraps the results of a function in an element]"""

    return_string = string or isinstance(element, str)

    if isinstance(element, str):
        from domonic.dom import Document

        element = Document.createElement(element).__class__

    def decorator(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            if result is None:
                result = ""
            if return_string is False:
                return element(result)
            else:
                return str(element(result))

        return wrapper

    return decorator


# @el(div)
# @el(span)

# def doctype(doctype):
#     """
#     @doctype('html')
#     @doctype('xhtml')
#     @doctype('xml')
#     """
#     def decorator(f):
#         @wraps(f)
#         def wrapper(*args, **kwargs):
#             return f(*args, **kwargs)
#         return wrapper
#     return decorator


def called(before=None, error: Callable[[Exception], None] = None):
    """Call a hook before a function runs, with light support for legacy callback sugar."""

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if before:
                _, _ = _invoke_before(before, f)
            try:
                return f(*args, **kwargs)
            except Exception as e:
                if error:
                    error(e)
                raise

        if before is None:
            wrapper()
            return wrapper

        params = inspect.signature(f).parameters
        if params:
            try:
                result, can_receive_value = _invoke_before(before, f)
                if can_receive_value:
                    f(result)
                else:
                    f()
            except Exception as e:
                if error:
                    error(e)
                else:
                    raise

        return wrapper

    return decorator


iife = called  # pass None for an iife

# def static(endpoint, update="11101"):
#     '''
#     render the endpoint to a cron timestamp. when user vists that function.
#     it will load the rendered version instead of executing the function.
#     '''
#     def dont_do_it(f):
#         return None
#     return dont_do_it


# https://www.python.org/dev/peps/pep-0318/
# https://stackoverflow.com/questions/15299878/how-to-use-python-decorators-to-check-function-arguments
def accepts(*types):
    def check_accepts(f):
        assert len(types) == f.__code__.co_argcount

        def new_f(*args, **kwds):
            for (a, t) in zip(args, types):
                assert isinstance(a, t), "arg %r does not match %s" % (a, t)
            return f(*args, **kwds)

        new_f.__name__ = f.__name__
        return new_f

    return check_accepts


# @accepts(int)


# CACHED = {}
# def memoize(f):
#     """
#     @memoize
#     def fib(n):
#         if n in (0, 1):
#             return n
#         return fib(n - 1) + fib(n - 2)
#     """
#     @functools.wraps(f)
#     def new_f(*args):
#         if args in CACHED:
#             return CACHED[args]
#         else:
#             result = f(*args)
#             CACHED[args] = result
#             return result
#     return new_f


# def requires_websocket(f): # decorate all functions that require websockets to warn user


def silence(*args, **kwargs):
    """Suppress stdout for the wrapped function.

    Supports both ``@silence`` and ``@silence()``.
    """

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*f_args, **f_kwargs):
            import contextlib

            with contextlib.redirect_stdout(io.StringIO()):
                return f(*f_args, **f_kwargs)

        return wrapper

    if args and callable(args[0]) and len(args) == 1 and not kwargs:
        return decorator(args[0])
    return decorator


def check(f):
    """Log entry and exit around a function call."""

    logger = logging.getLogger(__name__)

    @functools.wraps(f)
    def new_f(*args, **kwargs):
        logger.info("Entering %s", f.__name__)
        result = f(*args, **kwargs)
        logger.info("Exited %s", f.__name__)
        return result

    return new_f


def log(logger, level="info"):
    """@log(logging.getLogger('main'), level='warning')"""

    def log_decorator(fn):
        @functools.wraps(fn)
        def wrapper(*a, **kwa):
            getattr(logger, level)(fn.__name__)
            return fn(*a, **kwa)

        return wrapper

    return log_decorator


def instead(somethingelse):
    """Return a fallback value when the wrapped function raises."""

    def decorator(f):
        @functools.wraps(f)
        def new_f(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception:
                return somethingelse

        return new_f

    return decorator


def deprecated(func):
    """
    marks a function as deprecated.
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        raise Warning("Call to deprecated function {}.".format(func.__name__))
        return func(*args, **kwargs)

    return new_func


def as_json(func):
    """decorate any function to return json instead of a python obj
    note - used by JSON.py
    """

    def JSON_decorator(*args, **kwargs):
        import json

        return json.dumps(func(*args, **kwargs))

    return JSON_decorator


'''
def aka(names):
    """ @aka(*mylist) """
    def aka_decorator(fn):
        @functools.wraps(fn)
        def wrapper(*a, **kwa):
            return fn(*a, **kwa)
        return wrapper
    return log_decorator
'''


'''
import functools

# def DIV(func):
#     @functools.wraps(func)
#     def wrapper(*args, **kwargs):
#         result = func(*args, **kwargs)
#         return div(result)
#     return wrapper

# Create dynamic decorators for each tag
def create_html_decorator(tag):
    def decorator(func=None, **kwargs):
        # If no function is provided, we have been called with keyword arguments
        if func is None:
            return functools.partial(decorator, **kwargs)

        # Convert underscores in attributes to valid HTML (e.g., _class to class)
        updated_kwargs = {key.lstrip('_'): value for key, value in kwargs.items()}

        # Define the wrapper function that is called when the decorated function is called
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Call the original function to get the content
            content = func(*args, **kwargs)

            # Generate attributes for the HTML tag
            attrs = ''
            for key, value in updated_kwargs.items():
                attrs += f' {key}="{value}"'

            # Return the HTML wrapped content
            return f"<{tag}{attrs}>{content}</{tag}>"

        return wrapper

    return decorator


# from domonic.html import html_tags -- circular reference

# Dynamically create decorators for all HTML tags
for tag in html_tags:
    globals()[tag.upper()] = create_html_decorator(tag)


# @DIV(_id="cool", _class"awesome")
# @SPAN
# def somecontent():
#     return 'content only'
'''



"""
    domonic.terminal
    ====================================
    - call command line functions in python 3
"""
import os
import subprocess

from domonic.javascript import window


class TerminalException(Exception):
    """raised if the terminal throws an exception"""

    def __init__(self, error, message="An error message was recieved from terminal"):
        print(error, message)
        self.message = message
        super().__init__(self.message)


class command:
    """wrapper class for all terminal commands"""

    @staticmethod
    def run(cmd: str):
        """run

        runs any command on the terminal

        Args:
            cmd (str): The command you want to run on the terminal

        Returns:
            str: the response as a string
        """
        returned_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
        return returned_output.decode("utf-8")

    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs
        self.params = "".join([each.__str__() for each in args])
        self.result = ""

        if not hasattr(self, "first_run"):
            self.first_run = True
        else:
            self.first_run = False

        self.run_command()

    def __call__(self, *args, **kwargs):
        self.run_command()
        return self.result

    def run_command(self):
        if hasattr(self, "wait"):
            self.has_wait = True
            if self.first_run is True:
                self.first_run = False
                return
        else:
            self.has_wait = False

        if hasattr(self, "iterable"):
            if self.first_run is True:
                self.first_run = False
                return

        try:
            cmd = f"{self.name} {self.params}"
            cmd = cmd.strip()

            """
            # TODO - after doing all this. think i may have decided to go for 2 base commands instead
            # one iterable and one normal. its still hidden from user as just a different inhereted command
            # but would keep both simpler and can have more types. so will have to refactor all this again
            # for now this behaves how i want despite the _new_ hack and double call on this command
            """
            if self.has_wait is not True:
                returned_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
                self.result = returned_output.decode("utf-8")
            else:
                self.result = "PING FAIL"

                def kill_switch(proc):
                    proc.kill()
                    return

                # try:
                proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                intId = window.setInterval(kill_switch, 3000, proc)
                self.result = proc.stdout.readlines()
                window.clearInterval(intId)
                # except subprocess.TimeoutExpired as e:
                # self.result = returned_output.decode("utf-8")
                # print('process ran too long')

        except subprocess.CalledProcessError as e:
            print(e.output)
            self.result = e.output
            raise TerminalException(e, self.result)

        return

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        if hasattr(instance, "iterable"):
            is_iterable = instance.iterable
            if is_iterable:
                return instance

        return str(instance.result)

    def __str__(self) -> str:
        return str(self.result)

    def __getitem__(self, index):
        return self.result.splitlines()[index]


class cd(command):
    """
    NOTE - 'cd' does not run on terminal
    - cd is pointless as session opens and closes
    - so is overridden to change dirs via pure python
    """

    def run_command(self):
        os.chdir(self.params)


# tested --
ls = type("ls", (command,), {"name": "ls", "iterable": True})
rm = type("rm", (command,), {"name": "rm"})
ln = type("ln", (command,), {"name": "ln"})
mkdir = type("mkdir", (command,), {"name": "mkdir"})
rmdir = type("rmdir", (command,), {"name": "rmdir"})
touch = type("touch", (command,), {"name": "touch"})
df = type("df", (command,), {"name": "df", "iterable": True})
du = type("du", (command,), {"name": "du", "iterable": True})
ps = type("ps", (command,), {"name": "ps", "iterable": True})
cat = type("cat", (command,), {"name": "cat", "iterable": True})
# --

mv = type("mv", (command,), {"name": "mv"})
cp = type("cp", (command,), {"name": "cp"})
rm = type("rm", (command,), {"name": "rm"})
chmod = type("chmod", (command,), {"name": "chmod"})
chown = type("chown", (command,), {"name": "chown"})

users = type("users", (command,), {"name": "users", "iterable": True})
useradd = type("useradd", (command,), {"name": "useradd"})
userdel = type("userdel", (command,), {"name": "userdel"})
groups = type("groups", (command,), {"name": "groups"})
groupadd = type("groupadd", (command,), {"name": "groupadd"})
groupdel = type("groupdel", (command,), {"name": "groupdel"})

uniq = type("uniq", (command,), {"name": "uniq"})
sort = type("sort", (command,), {"name": "sort"})
diff = type("diff", (command,), {"name": "diff"})
# cmp = type('cmp', (command,), {'name': 'cmp'})

ssh = type("ssh", (command,), {"name": "ssh"})
scp = type("scp", (command,), {"name": "scp"})

more = type("more", (command,), {"name": "more"})
less = type("less", (command,), {"name": "less"})
# vi = type('vi', (command,), {'name': 'vi'})
# emacs = type('emacs', (command,), {'name': 'emacs'})
head = type("head", (command,), {"name": "head", "iterable": True})
tail = type("tail", (command,), {"name": "tail", "iterable": True})

uptime = type("uptime", (command,), {"name": "uptime"})
bash = type("bash", (command,), {"name": "bash"})
shutdown = type("shutdown", (command,), {"name": "shutdown"})
reboot = type("reboot", (command,), {"name": "reboot"})
# open = type('open', (command,), {'name': 'open'})
killall = type("killall", (command,), {"name": "killall"})
mkfile = type("mkfile", (command,), {"name": "mkfile"})
ifconfig = type("ifconfig", (command,), {"name": "ifconfig", "iterable": True})
ipconfig = type("ipconfig", (command,), {"name": "ipconfig", "iterable": True})
finger = type("finger", (command,), {"name": "finger"})
passwd = type("passwd", (command,), {"name": "passwd"})
whoami = type("whoami", (command,), {"name": "whoami"})
# help = type('help', (command,), {'name': 'help'})


class history(command):
    def run_command(self):
        self.iterable = True
        try:
            # history is empty for non interactive sessions so read file
            from os.path import expanduser, join

            with open(join(expanduser("~"), ".bash_history"), "r") as f:
                self.result = "".join(f.readlines())
        except Exception as e:
            print("failed to get history:", e)
            self.result = ""


ping = type("ping", (command,), {"name": "ping", "wait": True, "iterable": True})  # < TODO - need to stream feedback
man = type("man", (command,), {"name": "man"})
find = type("find", (command,), {"name": "find"})
awk = type("awk", (command,), {"name": "awk"})
grep = type("grep", (command,), {"name": "grep"})
cut = type("cut", (command,), {"name": "cut"})
sed = type("sed", (command,), {"name": "sed"})
apt = type("apt", (command,), {"name": "apt"})
gunzip = type("gunzip", (command,), {"name": "gunzip"})
tar = type("tar", (command,), {"name": "tar"})
gzip = type("gzip", (command,), {"name": "gzip"})
cron = type("cron", (command,), {"name": "cron"})
crontab = type("crontab", (command,), {"name": "crontab"})
xargs = type("xargs", (command,), {"name": "xargs"})
nautilus = type("nautilus", (command,), {"name": "nautilus"})  # Linux only?
date = type("date", (command,), {"name": "date"})
cal = type("cal", (command,), {"name": "cal"})
bc = type("bc", (command,), {"name": "bc"})

rsync = type("rsync", (command,), {"name": "rsync"})

# https://github.com/agarrharr/awesome-cli-apps
# - NOTE - these ones may need installing if you haven't already
git = type("git", (command,), {"name": "git"})
wget = type("wget", (command,), {"name": "wget"})
nmap = type("nmap", (command,), {"name": "nmap"})
nohup = type("nohup", (command,), {"name": "nohup"})
python = type("python", (command,), {"name": "python"})
npm = type("npm", (command,), {"name": "npm"})
cowsay = type("cowsay", (command,), {"name": "cowsay"})
pip = type("pip", (command,), {"name": "pip"})  # TODO - change to 'python3 -m pip'? # TODO - both iterable and not
say = type("say", (command,), {"name": "say"})  # Mac only?
gcc = type("gcc", (command,), {"name": "gcc"})
jq = type("jq", (command,), {"name": "jq"})
curl = type("curl", (command,), {"name": "curl"})
ffmpeg = type("ffmpeg", (command,), {"name": "ffmpeg"})
convert = type("convert", (command,), {"name": "convert"})  # imagemagik
figlet = type("figlet", (command,), {"name": "figlet"})
banner = type("banner", (command,), {"name": "banner"})
# imagemagik
# asciiview
# toilet
# figlet
# youtube-dl = type('youtube-dl', (command,), {'name': 'youtube-dl'}) # TODO - illegal chars
# tree = type('tree', (command,), {'name': 'tree'})
# vim = type('vim', (command,), {'name': 'vim'})
# mail = type('mail', (command,), {'name': 'mail'})
# vncserver = type('vncserver', (command,), {'name': 'vncserver'})
# tmux

# https://en.wikipedia.org/wiki/List_of_Unix_commands
# admin = type('admin', (command,), {'name': 'admin'})
ar = type("ar", (command,), {"name": "ar"})
asa = type("asa", (command,), {"name": "asa"})
at = type("at", (command,), {"name": "at"})
basename = type("basename", (command,), {"name": "basename"})
batch = type("batch", (command,), {"name": "batch"})
bg = type("bg", (command,), {"name": "bg"})
cc = type("cc", (command,), {"name": "cc"})
cflow = type("cflow", (command,), {"name": "cflow"})
chgrp = type("chgrp", (command,), {"name": "chgrp"})
cksum = type("cksum", (command,), {"name": "cksum"})
comm = type("comm", (command,), {"name": "comm"})
compress = type("compress", (command,), {"name": "compress"})
csplit = type("csplit", (command,), {"name": "csplit"})
ctags = type("ctags", (command,), {"name": "ctags"})
cxref = type("cxref", (command,), {"name": "cxref"})
dd = type("dd", (command,), {"name": "dd"})
delta = type("delta", (command,), {"name": "delta"})
dirname = type("dirname", (command,), {"name": "dirname"})
ed = type("ed", (command,), {"name": "ed"})
env = type("env", (command,), {"name": "env"})
ex = type("ex", (command,), {"name": "ex"})
expand = type("expand", (command,), {"name": "expand"})
expr = type("expr", (command,), {"name": "expr"})
fc = type("fc", (command,), {"name": "fc"})
fg = type("fg", (command,), {"name": "fg"})
file = type("file", (command,), {"name": "file"})
fold = type("fold", (command,), {"name": "fold"})
fort77 = type("fort77", (command,), {"name": "fort77"})
fuser = type("fuser", (command,), {"name": "fuser"})
gencat = type("gencat", (command,), {"name": "gencat"})
get = type("get", (command,), {"name": "get"})
getconf = type("getconf", (command,), {"name": "getconf"})
getopts = type("getopts", (command,), {"name": "getopts"})
iconv = type("iconv", (command,), {"name": "iconv"})
# id = type('id', (command,), {'name': 'id'})
ipcrm = type("ipcrm", (command,), {"name": "ipcrm"})
ipcs = type("ipcs", (command,), {"name": "ipcs"})
join = type("join", (command,), {"name": "join"})
lex = type("lex", (command,), {"name": "lex"})
link = type("link", (command,), {"name": "link"})
locale = type("locale", (command,), {"name": "locale"})
localedef = type("localedef", (command,), {"name": "localedef"})
logger = type("logger", (command,), {"name": "logger"})
logname = type("logname", (command,), {"name": "logname"})
lp = type("lp", (command,), {"name": "lp"})
m4 = type("m4", (command,), {"name": "m4"})
mailx = type("mailx", (command,), {"name": "mailx"})
make = type("make", (command,), {"name": "make"})
mesg = type("mesg", (command,), {"name": "mesg"})
mkdir = type("mkdir", (command,), {"name": "mkdir"})
mkfifo = type("mkfifo", (command,), {"name": "mkfifo"})
more = type("more", (command,), {"name": "more"})
newgrp = type("newgrp", (command,), {"name": "newgrp"})
nice = type("nice", (command,), {"name": "nice"})
nl = type("nl", (command,), {"name": "nl"})
nm = type("nm", (command,), {"name": "nm"})
od = type("od", (command,), {"name": "od"})
paste = type("paste", (command,), {"name": "paste"})
patch = type("patch", (command,), {"name": "patch"})
pathchk = type("pathchk", (command,), {"name": "pathchk"})
pax = type("pax", (command,), {"name": "pax"})
pr = type("pr", (command,), {"name": "pr"})
prs = type("prs", (command,), {"name": "prs"})
qalter = type("qalter", (command,), {"name": "qalter"})
qdel = type("qdel", (command,), {"name": "qdel"})
qhold = type("qhold", (command,), {"name": "qhold"})
qmove = type("qmove", (command,), {"name": "qmove"})
qmsg = type("qmsg", (command,), {"name": "qmsg"})
qrerun = type("qrerun", (command,), {"name": "qrerun"})
qrls = type("qrls", (command,), {"name": "qrls"})
qselect = type("qselect", (command,), {"name": "qselect"})
qsig = type("qsig", (command,), {"name": "qsig"})
qstat = type("qstat", (command,), {"name": "qstat"})
qsub = type("qsub", (command,), {"name": "qsub"})
read = type("read", (command,), {"name": "read"})
renice = type("renice", (command,), {"name": "renice"})
rmdel = type("rmdel", (command,), {"name": "rmdel"})
sact = type("sact", (command,), {"name": "sact"})
sccs = type("sccs", (command,), {"name": "sccs"})
sed = type("sed", (command,), {"name": "sed"})
sh = type("sh", (command,), {"name": "sh"})
sleep = type("sleep", (command,), {"name": "sleep"})
split = type("split", (command,), {"name": "split"})
strings = type("strings", (command,), {"name": "strings"})
strip = type("strip", (command,), {"name": "strip"})
stty = type("stty", (command,), {"name": "stty"})
tabs = type("tabs", (command,), {"name": "tabs"})
talk = type("talk", (command,), {"name": "talk"})
tee = type("tee", (command,), {"name": "tee"})
test = type("test", (command,), {"name": "test"})
time = type("time", (command,), {"name": "time"})
tput = type("tput", (command,), {"name": "tput"})
tr = type("tr", (command,), {"name": "tr"})
true = type("true", (command,), {"name": "true"})
tsort = type("tsort", (command,), {"name": "tsort"})
tty = type("tty", (command,), {"name": "tty"})
uname = type("uname", (command,), {"name": "uname"})
uncompress = type("uncompress", (command,), {"name": "uncompress"})
unexpand = type("unexpand", (command,), {"name": "unexpand"})
unget = type("unget", (command,), {"name": "unget"})
unlink = type("unlink", (command,), {"name": "unlink"})
uucp = type("uucp", (command,), {"name": "uucp"})
uudecode = type("uudecode", (command,), {"name": "uudecode"})
uuencode = type("uuencode", (command,), {"name": "uuencode"})
uustat = type("uustat", (command,), {"name": "uustat"})
uux = type("uux", (command,), {"name": "uux"})
val = type("val", (command,), {"name": "val"})
wc = type("wc", (command,), {"name": "wc"})  # < TODO - stream. plus figure out piping
what = type("what", (command,), {"name": "what"})
who = type("who", (command,), {"name": "who"})
write = type("write", (command,), {"name": "write"})
yacc = type("yacc", (command,), {"name": "yacc"})
zcat = type("zcat", (command,), {"name": "zcat"})

# bash builtins
alias = type("alias", (command,), {"name": "alias"})
bg = type("bg", (command,), {"name": "bg"})
bind = type("bind", (command,), {"name": "bind"})
# break = type('break', (command,), {'name': 'break'})
builtin = type("builtin", (command,), {"name": "builtin"})
caller = type("caller", (command,), {"name": "caller"})
# command = type('command', (command,), {'name': 'command'})
compgen = type("compgen", (command,), {"name": "compgen"})
complete = type("complete", (command,), {"name": "complete"})
compopt = type("compopt", (command,), {"name": "compopt"})
# continue = type('continue', (command,), {'name': 'continue'})
declare = type("declare", (command,), {"name": "declare"})
dirs = type("dirs", (command,), {"name": "dirs"})
disown = type("disown", (command,), {"name": "disown"})
echo = type("echo", (command,), {"name": "echo"})
enable = type("enable", (command,), {"name": "enable"})
# eval = type('eval', (command,), {'name': 'eval'})
# exec = type('exec', (command,), {'name': 'exec'})
exit = type("exit", (command,), {"name": "exit"})
export = type("export", (command,), {"name": "export"})
# false = type('false', (command,), {'name': 'false'})
fc = type("fc", (command,), {"name": "fc"})
fg = type("fg", (command,), {"name": "fg"})
getopts = type("getopts", (command,), {"name": "getopts"})
# hash = type('hash', (command,), {'name': 'hash'})
# help = type('help', (command,), {'name': 'help'})

jobs = type("jobs", (command,), {"name": "jobs"})
kill = type("kill", (command,), {"name": "kill"})
let = type("let", (command,), {"name": "let"})
local = type("local", (command,), {"name": "local"})
logout = type("logout", (command,), {"name": "logout"})
mapfile = type("mapfile", (command,), {"name": "mapfile"})
popd = type("popd", (command,), {"name": "popd"})
printf = type("printf", (command,), {"name": "printf"})
pushd = type("pushd", (command,), {"name": "pushd"})
pwd = type("pwd", (command,), {"name": "pwd"})
read = type("read", (command,), {"name": "read"})
readarray = type("readarray", (command,), {"name": "readarray"})
readonly = type("readonly", (command,), {"name": "readonly"})
# return = type('return', (command,), {'name': 'return'})
# set = type('set', (command,), {'name': 'set'})
shift = type("shift", (command,), {"name": "shift"})
shopt = type("shopt", (command,), {"name": "shopt"})
source = type("source", (command,), {"name": "source"})
suspend = type("suspend", (command,), {"name": "suspend"})
# test = type('test', (command,), {'name': 'test'})
times = type("times", (command,), {"name": "times"})
trap = type("trap", (command,), {"name": "trap"})
# true = type('true', (command,), {'name': 'true'})
# type = type('type', (command,), {'name': 'type'})
typeset = type("typeset", (command,), {"name": "typeset"})
ulimit = type("ulimit", (command,), {"name": "ulimit"})
umask = type("umask", (command,), {"name": "umask"})
unalias = type("unalias", (command,), {"name": "unalias"})
unset = type("unset", (command,), {"name": "unset"})
wait = type("wait", (command,), {"name": "wait"})


# https://ss64.com/bash/
# PIPE etc?



"""
    domonic.ext.lxml_html_
    ====================================

    Adapter for using lxml.html as a frontend parser while rebuilding a
    domonic document tree from the parsed lxml tree.
"""

from __future__ import annotations

from lxml import html

from domonic.ext.html5_parser_.dom import adapt


def parse(source, return_root=True, **kwargs):
    document = html.document_fromstring(source)
    tree = document.getroottree()
    return adapt(tree, return_root=return_root, **kwargs)



"""
    test_domonic
    ~~~~~~~~~~~~
    domonic core tests for __init__.py at the root of the domonic package

    (previously no tests. was using examples/parsing/page.py to drive dev)

"""

import unittest

from domonic import domonic


class TestCase(unittest.TestCase):
    """Tests for the domonic"""

    def test_load(self):
        t1 = domonic.load("<html></html>")
        print("test_load:::", t1, type(t1))

    def test_loads(self):
        # t1 = domonic.loads('<html></html>')
        # print(t1)
        pass

    def parse(self):
        t1 = domonic.parse("<html></html>")
        assert t1 == "html(),"  # hmm wondering if parse is correct term. as returns pyml strings

    def evaluate(self):
        t1 = domonic.evaluate("<html></html>")
        print(t1)

    def test_hacked_expat_parser(self):
        # test the  hacked version of the xpat parser
        print("test!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
        # t1 = domonic.parseString('<html></html>')
        # print(t1)
        t1 = domonic.parseString(
            """<html><head><link rel="stylesheet" href="https://unpkg.com/marx-css/css/marx.min.css" /><script src="https://code.jquery.com/jquery-3.5.1.min.js"></script><script>
	function add(){
		$('#results').html( Number($('#a').val()) + Number($('#b').val()) )};
</script></head><body><article><div><label>Add numbers:</label><input id="a" /><span>+</span><input id="b" /><button id="calculate_button" onclick="add();">Calculate</button><div>Result:<div id="results"></div></div></div></article></body></html>"""
        )
        print("RES:", t1)
        print("RES:", type(t1))
        print("RES:", str(t1))
        print(t1.getElementById("a"))

        # print(t1)
        return
        #         return

        # print( ':FIRE:', type(t1))
        # return
        # print(str(t1))
        # from domonic import render
        # print( render( t1 ) )

        print("test222!")
        t1 = domonic.parseString("<div></div>")
        print(t1)
        print(str(t1))
        # from domonic import render
        # print( render( t1 ) )

    def test_parse_string_with_expat_parser_option(self):
        page = domonic.parseString("<div><span>Hello</span></div>", parser="expat")
        self.assertIsNotNone(page)
        self.assertEqual(page.querySelector("span").text, "Hello")

    def test_parse_string_rejects_unknown_parser(self):
        with self.assertRaises(ValueError):
            domonic.parseString("<div></div>", parser="nope")

    def test_parse_string_with_html5_parser_option(self):
        try:
            page = domonic.parseString("<html><body><p>Hi</p></body></html>", parser="html5_parser")
        except ImportError:
            self.skipTest("html5_parser is not installed")
        else:
            self.assertIsNotNone(page)
            self.assertEqual(page.querySelector("p").text, "Hi")

    def test_parse_string_with_lxml_html_option(self):
        try:
            page = domonic.parseString("<html><body><p>Hi</p></body></html>", parser="lxml_html")
        except ImportError:
            self.skipTest("lxml is not installed")
        else:
            self.assertIsNotNone(page)
            self.assertEqual(page.querySelector("p").text, "Hi")

    def test_default_parser_setter(self):
        previous = domonic.get_default_parser()
        try:
            domonic.set_default_parser("expat")
            self.assertEqual(domonic.get_default_parser(), "expat")
            page = domonic.parseString("<div><span>Hello</span></div>")
            self.assertEqual(page.querySelector("span").text, "Hello")
        finally:
            domonic.set_default_parser(previous)

    def test_parse_string_with_markupever_option(self):
        try:
            page = domonic.parseString("<html><body><p>Hi</p></body></html>", parser="markupever")
        except ImportError:
            self.skipTest("markupever is not installed")
        else:
            self.assertIsNotNone(page)
            self.assertEqual(page.querySelector("p").text, "Hi")

    def test_parse_string_with_selectolax_option(self):
        try:
            page = domonic.parseString("<html><body><p>Hi</p></body></html>", parser="selectolax")
        except ImportError:
            self.skipTest("selectolax is not installed")
        else:
            self.assertIsNotNone(page)
            self.assertEqual(page.querySelector("p").text, "Hi")

    def test_parse_string_with_justhtml_option(self):
        try:
            page = domonic.parseString("<html><body><p>Hi</p></body></html>", parser="justhtml")
        except ImportError:
            self.skipTest("justhtml is not installed")
        else:
            self.assertIsNotNone(page)
            self.assertEqual(page.querySelector("p").text, "Hi")


if __name__ == "__main__":
    unittest.main()



"""
    domonic.webapi.xpath
    ====================================
    https://developer.mozilla.org/en-US/docs/Glossary/XPath

    uses elementpath lib.

    TODO - content strings must be TextNodes for it to work.
        so will have to iterate and update them. i.e. Treewalker

"""
from __future__ import annotations

from typing import Any

try:
    import elementpath
except ImportError:  # pragma: no cover - optional dependency
    elementpath = None


class XPathEvaluator:
    def __init__(self) -> None:
        pass

    def createExpression(self, expression: str) -> XPathExpression:  # , namespaces: Dict[str, str]) -> None:
        return XPathExpression(expression)


class XPathException:
    def __init__(self) -> None:
        pass


class XPathExpression:
    def __init__(self, expr: str):  # , resolver):
        # TODO - hack.
        # need to allow non underscore accessors to get underscored.
        # when that's fixed can remove this.
        expr = expr.replace("[@", "[@_")
        expr = expr.replace("[@__", "[@_")
        expr = expr.replace("(@", "(@_")
        expr = expr.replace("(@__", "(@_")
        expr = expr.replace("/@", "/@_")
        expr = expr.replace("/@__", "/@_")

        if len(expr) <= 0:
            raise Exception("no expression")
        self.expr = expr
        self.selector = elementpath.Selector(expr) if elementpath is not None else None

    # TODO - DRY - make some utils . just stole this from Treewalker.
    @staticmethod
    def _upgrade_dom(node: Any) -> Any:
        def upgrade(el):
            from domonic.dom import Text

            if isinstance(el, (Text, str)):
                return
            for child in el:
                if isinstance(child, str):
                    newchild = Text(child)
                    el.replaceChild(newchild, child)
                    newchild.parentNode = el

        node._iterate(node, upgrade)
        return node

    @staticmethod
    def _iter_descendants(node: Any) -> list[Any]:
        from domonic.dom import Element, Text

        descendants = []

        def walk(current):
            if isinstance(current, (Element, Text)):
                descendants.append(current)
            for child in getattr(current, "args", []):
                if isinstance(child, str):
                    continue
                walk(child)

        for child in getattr(node, "args", []):
            if isinstance(child, str):
                continue
            walk(child)
        return descendants

    @staticmethod
    def _iter_children(node: Any) -> list[Any]:
        from domonic.dom import Element, Text

        children = []
        for child in getattr(node, "args", []):
            if isinstance(child, (Element, Text)):
                children.append(child)
        return children

    @staticmethod
    def _parse_steps(expr: str) -> list[tuple[str, str]]:
        steps = []
        i = 0
        axis = "child"
        if expr.startswith("//"):
            axis = "descendant"
            i = 2
        elif expr.startswith("/"):
            axis = "child"
            i = 1
        current = []
        bracket_depth = 0
        paren_depth = 0
        quote = None
        while i < len(expr):
            char = expr[i]
            if quote is not None:
                current.append(char)
                if char == quote:
                    quote = None
                i += 1
                continue
            if char in ("'", '"'):
                quote = char
                current.append(char)
                i += 1
                continue
            if char == "[":
                bracket_depth += 1
            elif char == "]":
                bracket_depth -= 1
            elif char == "(":
                paren_depth += 1
            elif char == ")":
                paren_depth -= 1
            if bracket_depth == 0 and paren_depth == 0:
                if expr.startswith("//", i):
                    if current:
                        steps.append((axis, "".join(current)))
                        current = []
                    axis = "descendant"
                    i += 2
                    continue
                if char == "/":
                    if current:
                        steps.append((axis, "".join(current)))
                        current = []
                    axis = "child"
                    i += 1
                    continue
            current.append(char)
            i += 1
        if current:
            steps.append((axis, "".join(current)))
        return steps

    @staticmethod
    def _parse_step(step: str) -> tuple[str, list[str]]:
        predicates = []
        name_chars = []
        i = 0
        while i < len(step) and step[i] != "[":
            name_chars.append(step[i])
            i += 1
        name = "".join(name_chars).strip() or "*"
        while i < len(step):
            if step[i] != "[":
                i += 1
                continue
            depth = 1
            i += 1
            start = i
            quote = None
            while i < len(step) and depth > 0:
                char = step[i]
                if quote is not None:
                    if char == quote:
                        quote = None
                    i += 1
                    continue
                if char in ("'", '"'):
                    quote = char
                elif char == "[":
                    depth += 1
                elif char == "]":
                    depth -= 1
                    if depth == 0:
                        predicates.append(step[start:i].strip())
                        i += 1
                        break
                i += 1
        return name, predicates

    @staticmethod
    def _node_name(node: Any) -> str:
        return getattr(node, "tagName", getattr(node, "name", ""))

    @staticmethod
    def _predicate_matches(node: Any, predicate: str, index: int, nodes: list[Any]) -> bool:
        from domonic.dom import Text

        if predicate == "last()":
            return index == len(nodes) - 1
        if predicate.startswith('name()='):
            expected = predicate.split("=", 1)[1].strip().strip("'\"")
            return XPathExpression._node_name(node) == expected
        if predicate.startswith("@"):
            attr, _, value = predicate.partition("=")
            attr = attr[1:]
            if value:
                value = value.strip().strip("'\"")
                return node.getAttribute(attr) == value
            return node.getAttribute(attr) is not None
        if predicate.startswith("contains(") and predicate.endswith(")"):
            inner = predicate[len("contains(") : -1]
            attr_expr, _, needle = inner.partition(",")
            attr_expr = attr_expr.strip()
            needle = needle.strip().strip("'\"")
            if attr_expr.startswith("@"):
                attr_value = node.getAttribute(attr_expr[1:]) or ""
                return needle in attr_value
        return False

    @staticmethod
    def _step_matches(node: Any, name: str) -> bool:
        from domonic.dom import Element, Text

        if name == "text()":
            return isinstance(node, Text)
        if not isinstance(node, Element):
            return False
        if name == "*":
            return True
        return XPathExpression._node_name(node) == name

    def _fallback_select(self, node: Any) -> list[Any]:
        from domonic.dom import Text

        if self.expr == "/":
            return [node]

        nodes = [node]
        for axis, raw_step in self._parse_steps(self.expr):
            name, predicates = self._parse_step(raw_step)
            candidates = []
            for current in nodes:
                if axis == "descendant":
                    scope = self._iter_descendants(current)
                else:
                    scope = self._iter_children(current)
                for candidate in scope:
                    if self._step_matches(candidate, name):
                        candidates.append(candidate)
            if predicates:
                filtered = []
                for idx, candidate in enumerate(candidates):
                    if all(self._predicate_matches(candidate, predicate, idx, candidates) for predicate in predicates):
                        filtered.append(candidate)
                candidates = filtered
            nodes = candidates
        return nodes

    def evaluate(self, node: Any, type: int = 6):  # XPathResult.ANY_TYPE):
        # note: otherwise would fail on regular text?
        node = XPathExpression._upgrade_dom(node)
        if self.selector is not None:
            return XPathResult(self.selector.select(node), type)
        return XPathResult(self._fallback_select(node), type)


class XPathNSResolver:
    def __init__(self) -> None:
        pass


class XPathResult:

    ANY_TYPE = 0
    NUMBER_TYPE = 1
    STRING_TYPE = 2
    BOOLEAN_TYPE = 3
    UNORDERED_NODE_ITERATOR_TYPE = 4
    ORDERED_NODE_ITERATOR_TYPE = 5
    UNORDERED_NODE_SNAPSHOT_TYPE = 6
    ORDERED_NODE_SNAPSHOT_TYPE = 7
    ANY_UNORDERED_NODE_TYPE = 8
    FIRST_ORDERED_NODE_TYPE = 9

    def __init__(self, value: Any, _type: int):
        if _type == XPathResult.ANY_TYPE:
            tov = type(value)
            if tov == "object":
                _type = self.UNORDERED_NODE_ITERATOR_TYPE
            if tov == "boolean":
                _type = self.BOOLEAN_TYPE
            if tov == "string":
                _type = self.STRING_TYPE
            if tov == "number":
                _type = self.NUMBER_TYPE

        if _type < self.NUMBER_TYPE or self.FIRST_ORDERED_NODE_TYPE < _type:
            raise Exception(f"unknown type: {_type}")

        self.resultType = _type

        if _type == self.NUMBER_TYPE:
            # self.numberValue=value.number() if getattr(value,'isNodeSet',None) else toNumber(value)
            if getattr(value, "isNodeSet", None):
                self.numberValue = value  # .number()
            else:
                self.numberValue = float(value)
        elif _type == self.STRING_TYPE:
            # self.stringValue=value.string() if getattr(value,'isNodeSet',None) else toString(value)
            if getattr(value, "isNodeSet", None):
                self.stringValue = value  # .string()
            else:
                self.stringValue = str(value)
        elif _type == self.BOOLEAN_TYPE:
            # self.booleanValue=value.bool() if getattr(value,'isNodeSet',None) else toBoolean(value)
            if getattr(value, "isNodeSet", None):
                self.booleanValue = value  # .bool()
            else:
                self.booleanValue = bool(value)
        elif _type == self.ANY_UNORDERED_NODE_TYPE or _type == self.FIRST_ORDERED_NODE_TYPE:
            self.singleNodeValue = value  # .first()
        else:
            self.nodes = value  # .list()
            self.snapshotLength = len(value)
            self.index = 0
            self.invalidIteratorState = False

    # def iterateNext(self):
    #     node = self.nodes[self.index]
    #     self.index += 1
    #     return node

    # def snapshotItem(self, i):
    #     return self.nodes[i]



"""
domonic.window
==============

The DOM-connected ``Window`` implementation for domonic.

Be mindful that domonic has two window-shaped surfaces:

1. ``domonic.javascript.Window`` for JavaScript-style globals and timer helpers
2. ``domonic.window.Window`` for a browsing-context style object connected to
   ``document``, history, storage, media queries, and custom elements
"""

from __future__ import annotations

import re
import sys
from typing import Any, Callable

from domonic import domonic
from domonic.dom import Document, Element, Location, document
from domonic.events import CloseEvent, Event, EventTarget, FocusEvent, HashChangeEvent, PopStateEvent
from domonic.javascript import Promise, Window as JavaScriptWindow
from domonic.webapi.console import Console
from domonic.webapi.credentials import CredentialsContainer
from domonic.webapi.geo import Geolocation
from domonic.webapi.history import History
from domonic.webapi.netinfo import NetworkInformation
from domonic.webapi.webstorage import Storage


class MediaQueryList(EventTarget):
    """Result object returned by ``Window.matchMedia()``.

    It exposes the media query string, the current match state, and ``change``
    listeners in the familiar browser style.
    """

    def __init__(self, media: str, *, width: int, height: int) -> None:
        super().__init__()
        self.media = media
        self.matches = self._evaluate(media, width=width, height=height)
        self.onchange: Callable[[Event], Any] | None = None

    @staticmethod
    def _evaluate(media: str, *, width: int, height: int) -> bool:
        if not media:
            return False
        text = media.strip().lower()
        if text in ("all", "screen"):
            return True

        checks: list[bool] = []
        for label, value in (("min-width", width), ("max-width", width), ("min-height", height), ("max-height", height)):
            match = re.search(rf"\({label}\s*:\s*(\d+)px\)", text)
            if not match:
                continue
            target = int(match.group(1))
            if label.startswith("min-"):
                checks.append(value >= target)
            else:
                checks.append(value <= target)

        orientation_match = re.search(r"\(orientation\s*:\s*(portrait|landscape)\)", text)
        if orientation_match:
            orientation = "landscape" if width >= height else "portrait"
            checks.append(orientation == orientation_match.group(1))

        return all(checks) if checks else False

    def addListener(self, callback: Callable[[Event], Any]) -> None:
        self.addEventListener("change", callback)

    def removeListener(self, callback: Callable[[Event], Any]) -> None:
        self.removeEventListener("change", callback)


class CustomElementRegistry:
    """Registry for defining and upgrading custom elements.

    Access this through ``window.customElements`` to register custom element
    classes, wait for definitions, and upgrade parsed trees against the
    registry.
    """

    def __init__(self) -> None:
        self.store: dict[str, type[Element]] = {}
        self._constructors: dict[type, str] = {}
        self._when_defined: dict[str, list[Promise]] = {}

    @staticmethod
    def _validate_name(name: str) -> str:
        normalized = str(name).strip().lower()
        if not normalized or "-" not in normalized:
            raise ValueError("Invalid custom element name. Must contain hyphen: " + str(name))
        if not re.fullmatch(r"[a-z][.0-9_a-z-]*-[.0-9_a-z-]*", normalized):
            raise ValueError("Invalid custom element name: " + str(name))
        if normalized in {
            "annotation-xml",
            "color-profile",
            "font-face",
            "font-face-src",
            "font-face-uri",
            "font-face-format",
            "font-face-name",
            "missing-glyph",
        }:
            raise ValueError("Reserved custom element name: " + normalized)
        return normalized

    @staticmethod
    def _coerce_constructor(name: str, constructor: Callable[..., Any], options: dict[str, Any] | None = None) -> type[Element]:
        if not isinstance(constructor, type):
            raise TypeError("constructor must be a class")
        if issubclass(constructor, Element):
            if getattr(constructor, "name", None) in (None, ""):
                constructor.name = name
            return constructor
        attrs = {"name": name}
        if options is not None and "extends" in options:
            attrs["extends"] = options["extends"]
        return type(name.replace("-", "_"), (constructor, Element), attrs)

    def define(self, name: str, constructor: Callable[..., Any], options: dict[str, Any] | None = None) -> type:
        """Defines a new custom element."""
        normalized = self._validate_name(name)
        if normalized in self.store:
            raise ValueError("Custom element already defined: " + normalized)
        if constructor in self._constructors:
            raise ValueError("Custom element constructor already defined: " + self._constructors[constructor])

        element_class = self._coerce_constructor(normalized, constructor, options)
        element_class.name = normalized
        if options is not None and "extends" in options:
            element_class.extends = options["extends"]
        self.store[normalized] = element_class
        self._constructors[constructor] = normalized
        for promise in self._when_defined.pop(normalized, []):
            promise.resolve(element_class)
        return element_class

    def get(self, name: str) -> type | None:
        """Returns the constructor for the named custom element, or None."""
        return self.store.get(str(name).strip().lower())

    def getName(self, constructor: type) -> str | None:
        return self._constructors.get(constructor)

    def _upgrade_element(self, element: Element) -> Element:
        name = str(getattr(element, "tagName", getattr(element, "name", ""))).strip().lower()
        constructor = self.store.get(name)
        if constructor is None or isinstance(element, constructor):
            return element
        old_document = element.ownerDocument if isinstance(element.ownerDocument, Document) else None
        element.__class__ = constructor
        element.name = name
        element._custom_element_name = name
        if hasattr(constructor, "observedAttributes"):
            element.observedAttributes = getattr(constructor, "observedAttributes")
        if isinstance(old_document, Document) and getattr(element, "isConnected", False):
            callback = getattr(element, "connectedCallback", None)
            if callable(callback) and not getattr(element, "_custom_element_connected", False):
                element._custom_element_connected = True
                callback()
        return element

    def upgrade(self, root: Element | None = None) -> Element | None:
        if root is None:
            return None
        self._upgrade_element(root)
        for child in getattr(root, "childNodes", []):
            if isinstance(child, Element):
                self.upgrade(child)
        return root

    def whenDefined(self, name: str) -> Promise:
        normalized = self._validate_name(name)
        constructor = self.store.get(normalized)
        promise = Promise()
        if constructor is not None:
            promise.resolve(constructor)
            return promise
        self._when_defined.setdefault(normalized, []).append(promise)
        return promise


class Navigator:
    """Minimal browsing-environment navigator object."""

    cookieEnabled = False
    appName = "domonic"

    def __init__(self, *args, **kwargs):
        self.connection: NetworkInformation = NetworkInformation()
        self.credentials: CredentialsContainer = CredentialsContainer()
        self.geolocation: Geolocation = Geolocation()
        self.hid = None
        self.keyboard = None
        self.locks = None
        self.mediaCapabilities = None
        self.mediaSession = None
        self.mediaDevices = None
        self.presentation = None
        self.serial = None
        self.serviceWorker = None
        self.storage = None
        self.vendor = None
        self.webdriver = None
        self.xr = None
        self.buildID = None
        self.contacts = None
        self._screen = Screen()

    @property
    def onLine(self) -> bool:
        return True

    @property
    def platform(self) -> str:
        """Returns the platform"""
        if "darwin" in sys.platform:
            return "mac"
        if "linux" in sys.platform:
            return "linux"
        if "win32" in sys.platform:
            return "windows"
        return "unknown"

    @property
    def product(self) -> str:
        """Returns the product name"""
        return self.appName

    @property
    def userAgent(self) -> str:
        """Returns the user-agent header sent by the browser Navigator"""
        return f"domonic/{self.appName} ({self.platform})"

    @property
    def deviceMemory(self) -> float:
        return 1

    @property
    def doNotTrack(self):
        return "unspecified"

    @property
    def hardwareConcurrency(self):
        return 1

    @property
    def maxTouchPoints(self):
        return 1

    @staticmethod
    def registerProtocolHandler(scheme, url, title):
        return None

    @staticmethod
    def requestMediaKeySystemAccess(keySystem, supportedConfigurations):
        return None

    def canShare(self):
        return False

    def clearAppBadge(self):
        return None

    def getBattery(self):
        return {"charging": False, "chargingTime": 0, "dischargingTime": 0, "level": 1.0}

    @property
    def javaEnabled(self):
        return False

    def vibrate(self, pattern):
        return False


class Screen(EventTarget):
    # https://developer.mozilla.org/en-US/docs/Web/API/Screen

    def __init__(self):
        super().__init__()
        self.availLeft = 0
        self.availTop = 0
        self.availHeight = 768
        self.availWidth = 1024
        self.colorDepth = 24
        self.height = 768
        self.left = 0
        self.pixelDepth = 24
        self.top = 0
        self.width = 1024
        self.orientation = None

    @property
    def screenLeft(self) -> int:
        return self.left

    @property
    def screenTop(self) -> int:
        return self.top


class Window(JavaScriptWindow, EventTarget):
    def __init__(self):
        EventTarget.__init__(self)
        self.customElements = CustomElementRegistry()
        self._localStorage: Storage = Storage()
        self._sessionStorage: Storage = Storage()
        self._navigator: Navigator = Navigator()
        self._screen: Screen = self._navigator._screen
        self._document: Document = document
        self._document.defaultView = self
        self._location: Location = Location("https://eventual.technology")
        self._document.URL = self._location.href
        self._console: Console = Console()
        self._history: History = History(self)
        self._closed: bool = False
        self._focused: bool = True
        self._name: str = ""
        self._default_status: str = ""
        JavaScriptWindow.__init__(self)

    @staticmethod
    def _normalize_url(value: str | Location) -> str:
        href = value.href if isinstance(value, Location) else str(value)
        if href and "://" not in href:
            href = "https://" + href
        return href

    def _set_document(self, doc: Document) -> Document:
        previous_document = getattr(self, "_document", None)
        self._document = doc
        self._document.defaultView = self
        self._document.URL = self._location.href
        if previous_document is not None:
            self._document.referrer = getattr(previous_document, "URL", "") or ""
        return self._document

    def _fetch_document(self, url: str) -> Document | None:
        try:
            import requests
        except ModuleNotFoundError:
            return None

        try:
            response = requests.get(url, timeout=5)
            response.raise_for_status()
        except Exception:
            return None

        page = domonic.parseString(response.text)
        if page is not None:
            page.URL = url
        return page

    @property
    def history(self) -> History:
        return self._history

    @property
    def console(self) -> Console:
        return self._console

    @property
    def localStorage(self) -> Storage:
        return self._localStorage

    @property
    def sessionStorage(self) -> Storage:
        return self._sessionStorage

    @property
    def document(self) -> Document:
        return self._document

    @document.setter
    def document(self, value: Document) -> None:
        self._set_document(value)

    @property
    def location(self) -> Location:
        return self._location

    @location.setter
    def location(self, value: str | Location) -> None:
        if value is None:
            return
        previous_href = self._location.href
        href = self._normalize_url(value)
        if getattr(self._history, "skip_update", False) is False:
            self._history._update(href)
        self._location = Location(href)
        self._document.URL = href
        self._document.referrer = previous_href or ""
        if previous_href != href and previous_href.split("#", 1)[0] == href.split("#", 1)[0]:
            self.dispatchEvent(HashChangeEvent("hashchange", {"oldURL": previous_href, "newURL": href}))

        loaded_document = self._fetch_document(href)
        if loaded_document is not None:
            self._set_document(loaded_document)

    def blur(self):
        self._focused = False
        self.dispatchEvent(FocusEvent("blur", {"bubbles": False, "cancelable": False, "relatedTarget": None}))
        return None

    @property
    def closed(self) -> bool:
        return self._closed

    def close(self):
        self._closed = True
        self.dispatchEvent(CloseEvent("close", {"bubbles": False, "cancelable": False, "code": 1000, "reason": "", "wasClean": True}))
        return None

    def confirm(self, message: str):
        return True

    @property
    def defaultStatus(self):
        return self._default_status

    @defaultStatus.setter
    def defaultStatus(self, value=None):
        self._default_status = "" if value is None else str(value)

    def focus(self):
        self._focused = True
        self.dispatchEvent(FocusEvent("focus", {"bubbles": False, "cancelable": False, "relatedTarget": None}))
        return None

    def frameElement(self):
        return None

    def getComputedStyle(self, el, pseudo=None):
        return getattr(el, "style", None)

    def getSelection(self):
        return self.document.getSelection()

    @property
    def innerHeight(self):
        return self._screen.height

    @property
    def innerWidth(self):
        return self._screen.width

    def matchMedia(self, media_query_list):
        return MediaQueryList(media_query_list, width=self.innerWidth, height=self.innerHeight)

    def moveBy(self, x: int, y: int):
        self._screen.left += x
        self._screen.top += y

    def moveTo(self, x: int, y: int):
        self._screen.left = x
        self._screen.top = y

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def navigator(self):
        return self._navigator

    @property
    def screen(self) -> Screen:
        return self._screen

    @property
    def screenLeft(self) -> int:
        return self._screen.screenLeft

    @property
    def screenTop(self) -> int:
        return self._screen.screenTop


window = Window()
alert = window.alert
confirm = window.confirm
prompt = window.prompt



"""
    test_geom
    ~~~~~~~~~~~~
"""

import unittest
from domonic.constants.color import Color
from domonic.geom.shape import Circle, Line, Rect, Shape
from domonic.geom.vec2 import vec2
from domonic.geom.vec3 import vec3

import unittest

# TODO - make own class instead of piggy backing here
class TestColor(unittest.TestCase):
    def test_color_from_vec3(self):
        from domonic.geom.vec3 import vec3
        c = Color(vec3(255, 0, 0))
        self.assertEqual((c.r, c.g, c.b, c.a), (255, 0, 0, 1))

    def test_color_from_vec4(self):
        from domonic.geom.vec4 import vec4
        c = Color(vec4(255, 0, 0, 0.5))
        self.assertEqual((c.r, c.g, c.b, c.a), (255, 0, 0, 0.5))

    def test_color_from_hex_full(self):
        c = Color("#ff0000")
        self.assertEqual((c.r, c.g, c.b, c.a), (255, 0, 0, 1))

    # def test_color_from_hex_short(self):
    #     c = Color("#f00")
    #     self.assertEqual((c.r, c.g, c.b, c.a), (255, 0, 0, 1))

    # def test_color_from_named_color(self):
    #     c = Color("blue")
    #     self.assertEqual((c.r, c.g, c.b, c.a), (0, 0, 255, 1))

    def test_color_from_rgb(self):
        c = Color(255, 0, 0)
        self.assertEqual((c.r, c.g, c.b, c.a), (255, 0, 0, 1))

    def test_color_from_rgba(self):
        c = Color(255, 0, 0, 0.5)
        self.assertEqual((c.r, c.g, c.b, c.a), (255, 0, 0, 0.5))

    def test_invalid_hex_color(self):
        with self.assertRaises(ValueError):
            Color("#12345")

    def test_invalid_named_color(self):
        with self.assertRaises(ValueError):
            Color("invalid_color_name")

    def test_invalid_numeric_inputs(self):
        with self.assertRaises(ValueError):
            Color(255, 255)  # Missing blue
        with self.assertRaises(ValueError):
            Color(255)       # Missing green and blue

    def test_color_alpha_default(self):
        c = Color(100, 150, 200)
        self.assertEqual(c.a, 1)  # Alpha defaults to 1

    def test_color_alpha_explicit(self):
        c = Color(100, 150, 200, 0.7)
        self.assertEqual(c.a, 0.7)

    # def test_color_repr(self):
    #     c = Color(255, 255, 255)
    #     self.assertEqual(str(c), "Color(255, 255, 255, 1)")

    def test_color_eq(self):
        c1 = Color(255, 255, 255)
        c2 = Color(255, 255, 255)
        c3 = Color(0, 0, 0)
        self.assertEqual(c1, c2)
        self.assertNotEqual(c1, c3)

    def test_hex_to_rgb(self):
        self.assertEqual(Color.hex2rgb("#ff0000"), (255, 0, 0))
        with self.assertRaises(ValueError):
            Color.hex2rgb("#12345")

    # def test_named_colors(self):
    #     named_colors = Color.named_colors()
    #     self.assertIn("red", named_colors)
    #     self.assertEqual(named_colors["red"], (255, 0, 0))
    #     self.assertEqual(named_colors["black"], (0, 0, 0))


class TestCase(unittest.TestCase):

    def test_vec2_operations(self):
        v = vec2(10, 10)
        self.assertEqual(v, v)  # Identity test
        self.assertEqual(v * 2, vec2(20, 20))
        self.assertEqual(v / 2, vec2(5, 5))
        self.assertEqual(v - 2, vec2(8, 8))
        self.assertEqual(v + v, vec2(20, 20))
        self.assertEqual(v * v, vec2(100, 100))
        self.assertEqual(v / v, vec2(1, 1))
        self.assertEqual(v - v, vec2(0, 0))
        self.assertEqual((v[0], v[1]), (10, 10))
        self.assertEqual((v["x"], v["y"]), (10, 10))
        self.assertEqual((v.x, v.y), (10, 10))

    def test_vec3_operations(self):
        v = vec3(10, 10, 10)
        self.assertEqual(v * 2, vec3(20, 20, 20))
        self.assertEqual(v / 2, vec3(5, 5, 5))
        self.assertEqual(v - 2, vec3(8, 8, 8))
        self.assertEqual(v + v, vec3(20, 20, 20))
        self.assertEqual(v * v, vec3(100, 100, 100))
        self.assertEqual(v - v, vec3(0, 0, 0))

    def test_shape_color(self):
        s = Shape(color=Color.red)
        self.assertEqual(s.color, Color.red)

    def test_rect_operations(self):
        r = Rect(0, 0, 10, 10)
        self.assertEqual(r * 10, Rect(0, 0, 100, 100))
        self.assertEqual(r / 10, Rect(0, 0, 1, 1))

        a = Rect(0, 0, 10, 10)
        b = Rect(0, 0, 20, 20)
        c = Rect(0, 0, 30, 30)
        self.assertEqual(a + b + c, Rect(0, 0, 60, 60))

    def test_line_creation(self):
        l = Line(vec2(0, 0), vec2(10, 10))
        self.assertEqual(l.start, vec2(0, 0))
        self.assertEqual(l.end, vec2(10, 10))

    def test_circle_operations(self):
        c = Circle(0, 0, 10)
        self.assertEqual(c.radius, 10)
        self.assertEqual(c * 4, Circle(0, 0, 40))

    def test_color_creation(self):
        c1 = Color("#ff00ff")
        self.assertEqual(str(c1), "#ff00ff")

        c3 = Color(255, 255, 255)
        self.assertEqual(str(c3), "#ffffff")

        c4 = Color(255, 255, 255, 0)
        self.assertEqual((c4.r, c4.g, c4.b, c4.a), (255, 255, 255, 0))


if __name__ == "__main__":
    unittest.main()



"""
    test_templates
    ~~~~~~~~~~~~
    tests for templates
"""

import unittest

from domonic.templates import (
    blank_page,
    maintenance_page,
    message_page,
    redirect_page,
    runtime_page,
    status_page,
)


class TestTemplates(unittest.TestCase):
    def test_status_page_whole_page(self):
        page = str(status_page(404))
        self.assertIn("<title>404</title>", page)
        self.assertIn("<h1>404</h1>", page)
        self.assertIn("Not Found", page)

    def test_status_page_fragment(self):
        page = str(status_page(401, False))
        stripped = page.lstrip()
        self.assertTrue(stripped.startswith("<body>") or stripped.startswith("<div"))
        self.assertIn("Unauthorized", page)

    def test_blank_page(self):
        page = str(blank_page("Hello", "Hi there"))
        self.assertIn("<title>Hello</title>", page)
        self.assertIn("Hi there", page)

    def test_message_page(self):
        page = str(message_page("Heads Up", "Notice", "Something happened", "Traceback info"))
        self.assertIn("<title>Heads Up</title>", page)
        self.assertIn("Something happened", page)
        self.assertIn("Traceback info", page)

    def test_redirect_page(self):
        page = str(redirect_page("/next", 5))
        self.assertIn('http_equiv="refresh"', page)
        self.assertIn('content="5;url=/next"', page)
        self.assertIn("Redirecting", page)

    def test_maintenance_page(self):
        page = str(maintenance_page("10 minutes"))
        self.assertIn("Service Unavailable", page)
        self.assertIn("Retry after: 10 minutes", page)

    def test_runtime_page(self):
        page = str(runtime_page())
        self.assertIn("Runtime Information", page)
        self.assertIn("Working Directory", page)
        self.assertIn("Process ID", page)

    def test_runtime_page_with_environment(self):
        page = str(runtime_page(include_environment=True))
        self.assertIn("Environment", page)
        self.assertIn('id="environment-info"', page)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

"""
domonic.animation
=================

Web Animations-style helpers for domonic.

This module provides a practical Python surface for working with animation
timing, keyframes, playback state, and DOM-connected ``Element.animate(...)``
behaviour.
"""

import math
import re
from dataclasses import dataclass
from typing import Any

from domonic.events import Event, EventTarget
from domonic.lerpy import lerp
from domonic.lerpy.easing import Linear, Quad


def _coerce_timing_options(options: EffectTiming | dict[str, Any] | None) -> dict[str, Any]:
    if options is None:
        return {}
    if isinstance(options, EffectTiming):
        return options.to_dict()
    if isinstance(options, dict):
        return dict(options)
    if isinstance(options, (int, float)):
        return {"duration": float(options)}
    raise TypeError("Unsupported timing options")


def _normalize_iteration_count(value: Any) -> float:
    if value in (None, "", "Infinity", "infinity"):
        return math.inf if value in ("Infinity", "infinity") else 1.0
    return float(value)


def _parse_interpolable(value: Any) -> tuple[float, str] | None:
    if isinstance(value, (int, float)):
        return float(value), ""
    if isinstance(value, str):
        match = re.fullmatch(r"\s*(-?\d+(?:\.\d+)?)([a-zA-Z%]*)\s*", value)
        if match:
            return float(match.group(1)), match.group(2)
    return None


def _format_interpolated(value: float, unit: str) -> str | float:
    if unit:
        formatted = int(value) if float(value).is_integer() else round(value, 6)
        return f"{formatted}{unit}"
    return int(value) if float(value).is_integer() else value


def _resolve_easing(name: str):
    easing = (name or "linear").strip().lower()
    mapping = {
        "linear": Linear.easeIn,
        "ease": Quad.easeInOut,
        "ease-in": Quad.easeIn,
        "ease-out": Quad.easeOut,
        "ease-in-out": Quad.easeInOut,
    }
    return mapping.get(easing, Linear.easeIn)


@dataclass
class EffectTiming:
    """Timing parameters shared by animation effects.

    This is the author-facing timing object that describes delay, duration,
    direction, fill mode, easing, and iteration behaviour.
    """

    delay: float = 0.0
    direction: str = "normal"
    duration: float = 0.0
    easing: str = "linear"
    endDelay: float = 0.0
    fill: str = "none"
    iterationStart: float = 0.0
    iterations: float = 1.0

    def __init__(
        self,
        delay: float = 0.0,
        direction: str = "normal",
        duration: float = 0.0,
        easing: str = "linear",
        endDelay: float = 0.0,
        fill: str = "none",
        iterationStart: float = 0.0,
        iterations: float = 1.0,
    ) -> None:
        self.delay = float(delay)
        self.direction = direction
        self.duration = float(duration)
        self.easing = easing
        self.endDelay = float(endDelay)
        self.fill = fill
        self.iterationStart = float(iterationStart)
        self.iterations = _normalize_iteration_count(iterations)

    def to_dict(self) -> dict[str, Any]:
        return {
            "delay": self.delay,
            "direction": self.direction,
            "duration": self.duration,
            "easing": self.easing,
            "endDelay": self.endDelay,
            "fill": self.fill,
            "iterationStart": self.iterationStart,
            "iterations": self.iterations,
        }


class ComputedEffectTiming(EffectTiming):
    """Resolved timing data derived from an effect plus a current local time."""

    def __init__(
        self,
        timing: EffectTiming | None = None,
        *,
        activeDuration: float = 0.0,
        currentIteration: int | None = None,
        endTime: float = 0.0,
        progress: float | None = None,
    ) -> None:
        timing = timing or EffectTiming()
        super().__init__(**timing.to_dict())
        self.activeDuration = activeDuration
        self.currentIteration = currentIteration
        self.endTime = endTime
        self.progress = progress


class AnimationPlaybackEvent(Event):
    """Event carrying playback timing information for animation lifecycle hooks."""

    def __init__(self, _type: str, options: dict[str, Any] | None = None, *args, **kwargs) -> None:
        options = options or kwargs
        self.currentTime = options.get("currentTime")
        self.timelineTime = options.get("timelineTime")
        super().__init__(_type, options, *args, **kwargs)


class AnimationEffect:
    """Base class for effects that can be sampled over time and applied to targets."""

    def __init__(self, timing: EffectTiming | dict[str, Any] | None = None) -> None:
        self._timing = EffectTiming(**_coerce_timing_options(timing))

    def getTiming(self) -> EffectTiming:
        return EffectTiming(**self._timing.to_dict())

    def updateTiming(self, options: EffectTiming | dict[str, Any] | None = None) -> None:
        if options is None:
            return
        updated = self._timing.to_dict()
        updated.update(_coerce_timing_options(options))
        self._timing = EffectTiming(**updated)

    def getComputedTiming(self, local_time: float | None = None) -> ComputedEffectTiming:
        timing = self.getTiming()
        iterations = timing.iterations
        active_duration = math.inf if math.isinf(iterations) else timing.duration * iterations
        end_time = timing.delay + active_duration + timing.endDelay
        progress = None
        current_iteration = None

        if local_time is not None:
            active_time = local_time - timing.delay
            if active_time < 0:
                if timing.fill in {"backwards", "both"}:
                    progress = 0.0
                    current_iteration = 0
            elif math.isinf(active_duration) or active_time < active_duration:
                duration = timing.duration if timing.duration > 0 else 1.0
                current_iteration = int(active_time // duration) if timing.duration > 0 else 0
                simple_progress = 1.0 if timing.duration == 0 else (active_time % duration) / duration
                progress = self._directed_progress(simple_progress, current_iteration)
            elif active_duration == 0 or active_time == active_duration:
                progress = self._directed_progress(1.0, max(0, int(iterations) - 1 if not math.isinf(iterations) else 0))
                current_iteration = None if math.isinf(iterations) else max(0, int(iterations) - 1)
            elif timing.fill in {"forwards", "both"}:
                progress = self._directed_progress(1.0, max(0, int(iterations) - 1 if not math.isinf(iterations) else 0))
                current_iteration = None if math.isinf(iterations) else max(0, int(iterations) - 1)

        return ComputedEffectTiming(
            timing,
            activeDuration=active_duration,
            currentIteration=current_iteration,
            endTime=end_time,
            progress=progress,
        )

    def _directed_progress(self, progress: float, iteration: int) -> float:
        direction = self._timing.direction
        if direction == "reverse":
            return 1.0 - progress
        if direction == "alternate" and iteration % 2 == 1:
            return 1.0 - progress
        if direction == "alternate-reverse":
            return progress if iteration % 2 == 1 else 1.0 - progress
        return progress

    def apply(self, local_time: float | None = None) -> None:
        raise NotImplementedError


class KeyframeEffect(AnimationEffect):
    """Keyframe-based effect bound to a target object or DOM element."""

    def __init__(
        self,
        target: Any,
        keyframes: list[dict[str, Any]] | dict[str, Any],
        options: EffectTiming | dict[str, Any] | None = None,
    ) -> None:
        self.target = target
        self._keyframes = self._normalize_keyframes(keyframes)
        super().__init__(options)

    @staticmethod
    def _normalize_keyframes(keyframes: list[dict[str, Any]] | dict[str, Any]) -> list[dict[str, Any]]:
        frames = [dict(frame) for frame in (keyframes if isinstance(keyframes, list) else [keyframes])]
        if not frames:
            return [{"offset": 0.0}, {"offset": 1.0}]

        for index, frame in enumerate(frames):
            if "offset" in frame and frame["offset"] is not None:
                frame["offset"] = float(frame["offset"])
            else:
                frame["offset"] = index / (len(frames) - 1) if len(frames) > 1 else 1.0
        frames.sort(key=lambda frame: frame["offset"])
        return frames

    def getKeyframes(self) -> list[dict[str, Any]]:
        return [dict(frame) for frame in self._keyframes]

    def _interpolate_value(self, prop: str, start: Any, end: Any, progress: float) -> Any:
        easing_fn = _resolve_easing(self._timing.easing)
        eased = easing_fn(progress, 0, 1, 1, 0, 0)
        start_value = _parse_interpolable(start)
        end_value = _parse_interpolable(end)
        if start_value is not None and end_value is not None and start_value[1] == end_value[1]:
            return _format_interpolated(lerp(start_value[0], end_value[0], eased), start_value[1])
        return end if eased >= 0.5 else start

    def _apply_property(self, prop: str, value: Any) -> None:
        if hasattr(self.target, "style") and hasattr(self.target.style, "__dict__"):
            setattr(self.target.style, prop, value)
            return
        if hasattr(self.target, prop):
            setattr(self.target, prop, value)
            return
        if hasattr(self.target, "setAttribute"):
            self.target.setAttribute(prop, value)

    def apply(self, local_time: float | None = None) -> None:
        computed = self.getComputedTiming(local_time)
        if computed.progress is None:
            return

        progress = computed.progress
        before = self._keyframes[0]
        after = self._keyframes[-1]
        for index in range(len(self._keyframes) - 1):
            start = self._keyframes[index]
            end = self._keyframes[index + 1]
            if start["offset"] <= progress <= end["offset"]:
                before = start
                after = end
                break

        span = after["offset"] - before["offset"]
        local_progress = 1.0 if span == 0 else (progress - before["offset"]) / span
        properties = set(before.keys()) | set(after.keys())
        properties.discard("offset")
        properties.discard("easing")
        properties.discard("composite")

        for prop in properties:
            start_value = before.get(prop, after.get(prop))
            end_value = after.get(prop, before.get(prop))
            value = self._interpolate_value(prop, start_value, end_value, local_progress)
            self._apply_property(prop, value)


class Animation(EventTarget):
    def __init__(
        self,
        effect: AnimationEffect | None = None,
        timeline: Any = None,
        *,
        id: str = "",
    ) -> None:
        super().__init__()
        self.effect = effect
        self.timeline = timeline
        self.id = id
        self.playbackRate: float = 1.0
        self.startTime: float | None = None
        self._currentTime: float | None = None
        self.playState: str = "idle"

    @property
    def currentTime(self) -> float | None:
        if self.playState == "running" and self.startTime is not None and self.timeline is not None:
            self._currentTime = (self.timeline.currentTime - self.startTime) * self.playbackRate
        return self._currentTime

    @currentTime.setter
    def currentTime(self, value: float | None) -> None:
        self._currentTime = None if value is None else float(value)
        if self.effect is not None and self._currentTime is not None:
            self.effect.apply(self._currentTime)

    def play(self) -> None:
        if self.timeline is not None:
            self.startTime = self.timeline.currentTime - (self._currentTime or 0.0)
        self.playState = "running"
        if self._currentTime is None:
            self.currentTime = 0.0

    def pause(self) -> None:
        self.currentTime = self.currentTime
        self.playState = "paused"

    def cancel(self) -> None:
        self.playState = "idle"
        self.startTime = None
        self._currentTime = None
        self.dispatchEvent(
            AnimationPlaybackEvent(
                "cancel",
                {"currentTime": None, "timelineTime": getattr(self.timeline, "currentTime", None)},
            )
        )

    def finish(self) -> None:
        effect_time = 0.0
        if self.effect is not None:
            effect_time = self.effect.getComputedTiming().activeDuration
        self.currentTime = effect_time
        self.playState = "finished"
        self.dispatchEvent(
            AnimationPlaybackEvent(
                "finish",
                {"currentTime": self.currentTime, "timelineTime": getattr(self.timeline, "currentTime", None)},
            )
        )

    def reverse(self) -> None:
        self.playbackRate = -self.playbackRate
        self.play()

    def updatePlaybackRate(self, playback_rate: float) -> None:
        self.currentTime = self.currentTime
        self.playbackRate = float(playback_rate)



#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: Apache 2.0 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from lxml.etree import _Comment

from domonic.dom import DOMImplementation

impl = DOMImplementation()

try:
    dict_items = dict.iteritems
except AttributeError:
    dict_items = dict.items

_ELEM_NAME_CACHE = {}


def elem_name_parts(elem):
    key = (elem.tag, elem.prefix)
    cached = _ELEM_NAME_CACHE.get(key)
    if cached is not None:
        return cached

    tag, prefix = key
    if tag.startswith("{"):
        uri, _, name = tag.rpartition("}")
        if prefix:
            name = prefix + ":" + name
        cached = (uri[1:], name)
    else:
        cached = (None, tag)

    _ELEM_NAME_CACHE[key] = cached
    return cached


def attr_name_parts(name, elem, val):
    if name.startswith("{"):
        uri, _, name = name.rpartition("}")
        uri = uri[1:]
        for prefix, quri in dict_items(elem.nsmap):
            if quri == uri:
                break
        else:
            prefix = None
        if prefix:
            name = prefix + ":" + name
        return uri, name, val
    return None, name, val


def add_namespace_declarations(src, dest):
    changed = src.nsmap
    if changed:
        p = src.getparent()
        if p is not None:
            # Only add namespace declarations different from the parent's
            p = p.nsmap or {}
            changed = {k: v for k, v in dict_items(changed) if v != p.get(k)}
        for prefix, uri in dict_items(changed):
            attr = ("xmlns:" + prefix) if prefix else "xmlns"
            dest.setAttributeNS("xmlns", attr, uri)


def adapt(source_tree, return_root=True, **kw):
    source_root = source_tree.getroot()
    uri, qname = elem_name_parts(source_root)
    dest_tree = impl.createDocument(uri, qname, None)
    dest_tree.doctype = source_tree.docinfo.doctype
    dest_root = dest_tree.documentElement
    create_text = dest_tree.createTextNode
    create_comment = dest_tree.createComment
    create_element_ns = dest_tree.createElementNS
    stack = [(source_root, dest_root)]
    while stack:
        src, dest = stack.pop()
        append_child = dest.appendChild
        set_attribute_ns = dest.setAttributeNS

        text = src.text
        if text:
            append_child(create_text(text))
        add_namespace_declarations(src, dest)
        for name, val in src.items():
            set_attribute_ns(*attr_name_parts(name, src, val))
        for child in src.iterchildren():
            if isinstance(child, _Comment):
                dchild = create_comment((child.text or "").replace("--", "—"))
            else:
                dchild = create_element_ns(*elem_name_parts(child))
                stack.append((child, dchild))
            append_child(dchild)
            tail = child.tail
            if tail:
                append_child(create_text(tail))

    if return_root or (qname and qname.lower() == "html"):
        return dest_root
    return dest_tree



"""
    test_window
    ~~~~~~~~~~~~~~~~
"""

import unittest

from domonic.html import body, div
from domonic.events import CloseEvent
from domonic.window import MediaQueryList, Window


class TestCase(unittest.TestCase):
    def test_window_core_properties(self):
        win = Window()

        self.assertIs(win.document.defaultView, win)
        self.assertEqual(win.location.href, "https://eventual.technology")
        self.assertEqual(win.innerWidth, win.screen.width)
        self.assertEqual(win.innerHeight, win.screen.height)
        self.assertFalse(win.closed)

    def test_window_focus_close_and_name(self):
        win = Window()
        events = []
        close_events = []

        win.addEventListener("focus", lambda event: events.append(event.type))
        win.addEventListener("blur", lambda event: events.append(event.type))
        win.addEventListener("close", lambda event: close_events.append(event))
        win.focus()
        win.blur()
        win.name = "main"
        win.close()

        self.assertEqual(events, ["focus", "blur"])
        self.assertEqual(len(close_events), 1)
        self.assertIsInstance(close_events[0], CloseEvent)
        self.assertEqual(close_events[0].code, 1000)
        self.assertTrue(close_events[0].wasClean)
        self.assertEqual(win.name, "main")
        self.assertTrue(win.closed)

    def test_window_computed_style_and_selection(self):
        win = Window()
        el = div()
        el.style.width = "120px"
        win.document.body = body(el)

        self.assertIs(win.getComputedStyle(el), el.style)
        self.assertIs(win.getSelection(), win.document.getSelection())

    def test_match_media_and_position_helpers(self):
        win = Window()
        query = win.matchMedia("(min-width: 800px)")

        self.assertIsInstance(query, MediaQueryList)
        self.assertTrue(query.matches)

        win.moveTo(5, 10)
        self.assertEqual((win.screenLeft, win.screenTop), (5, 10))

        win.moveBy(3, 4)
        self.assertEqual((win.screenLeft, win.screenTop), (8, 14))

    def test_location_assignment_without_network(self):
        win = Window()

        win.location = "example.com"

        self.assertEqual(win.location.href, "https://example.com")
        self.assertEqual(win.document.URL, "https://example.com")
        self.assertEqual(win.document.referrer, "https://eventual.technology")
        self.assertEqual(win.history.state, "https://example.com")

    def test_document_metadata_properties_are_window_backed(self):
        win = Window()

        self.assertIs(win.document.defaultView, win)
        self.assertEqual(win.document.designMode, "off")

        win.document.designMode = "on"
        self.assertEqual(win.document.designMode, "on")

        win.document.cookie = "session=abc123"
        win.document.cookie = "theme=dark"
        self.assertIn("session=abc123", win.document.cookie)
        self.assertIn("theme=dark", win.document.cookie)

    def test_hashchange_and_popstate_events(self):
        win = Window()
        events = []

        win.addEventListener("hashchange", lambda event: events.append((event.type, event.oldURL, event.newURL)))
        win.addEventListener("popstate", lambda event: events.append((event.type, event.state)))

        win.location = "https://example.com#one"
        win.location = "https://example.com#two"
        win.history.back()

        self.assertIn(("hashchange", "https://example.com#one", "https://example.com#two"), events)
        self.assertIn(("popstate", "https://example.com#one"), events)

    def test_navigator_basic_specish_helpers(self):
        win = Window()

        self.assertEqual(win.navigator.registerProtocolHandler("mailto", "/compose", "Mail"), None)
        self.assertEqual(win.navigator.requestMediaKeySystemAccess("org.example", []), None)
        self.assertEqual(win.navigator.clearAppBadge(), None)
        self.assertEqual(win.navigator.getBattery()["level"], 1.0)
        self.assertFalse(win.navigator.vibrate([100]))


if __name__ == "__main__":
    unittest.main()



"""
    domonic.constants.entities
    ====================================

"""

import html


class Entity:

    # __slots__ = ('entity',)

    def __init__(self, entity: str) -> None:
        self.entity = entity

    def __str__(self) -> str:
        return html.unescape(self.entity)

    def __repr__(self) -> str:
        return f"Entity({self.entity!r})"


class Char:
    def __init__(self, character: str) -> None:
        self.character = character

    def __str__(self) -> str:
        return html.escape(self.character)

    # def __repr__(self):
    #     return self.character

    # web
    # ASCII Characters (Printable)
    SPACE: str = "&#32;"
    EXCLAMATION_MARK: str = "&#33;"  #: !
    QUOTATION_MARK: str = "&#34;"  #: "
    NUMBER_SIGN: str = "&#35;"  #: #
    DOLLAR_SIGN: str = "&#36;"  #: $
    PERCENT_SIGN: str = "&#37;"  #: %
    AMPERSAND: str = "&amp;"  #: &
    APOSTROPHE: str = "&#39;"  #: '
    OPENING_PARENTHESIS: str = "&#40;"  #: (
    LEFT_PARENTHESIS: str = "&#40;"  #: (
    CLOSING_PARENTHESIS: str = "&#41;"  #: )
    RIGHT_PARENTHESIS: str = "&#41;"  #: )
    ASTERISK: str = "&#42;"  #: *
    PLUS_SIGN: str = "&#43;"  #: +
    COMMA: str = "&#44;"  #: ,
    HYPHEN: str = "&#45;"  #: -
    PERIOD: str = "&#46;"  #: .
    SLASH: str = "&#47;"  #: /
    ZERO: str = "&#48;"  #: 0
    ONE: str = "&#49;"  #: 1
    TWO: str = "&#50;"  #: 2
    THREE: str = "&#51;"  #: 3
    FOUR: str = "&#52;"  #: 4
    FIVE: str = "&#53;"  #: 5
    SIX: str = "&#54;"  #: 6
    SEVEN: str = "&#55;"  #: 7
    EIGHT: str = "&#56;"  #: 8
    NINE: str = "&#57;"  #: 9

    COLON: str = "&#58;"  #: :
    SEMICOLON: str = "&#59;"  #: ;
    LESS_THAN: str = "&lt;"  #: <
    EQUALS_SIGN: str = "&#61;"  # :: str =
    GREATER_THAN: str = "&gt;"  #: >
    QUESTION_MARK: str = "&#63;"  #: ?
    AT_SIGN: str = "&#64;"  #: @

    UPPERCASE_A: str = "&#65;"  #: A
    UPPERCASE_B: str = "&#66;"  #: B
    UPPERCASE_C: str = "&#67;"  #: C
    UPPERCASE_D: str = "&#68;"  #: D
    UPPERCASE_E: str = "&#69;"  #: E
    UPPERCASE_F: str = "&#70;"  #: F
    UPPERCASE_G: str = "&#71;"  #: G
    UPPERCASE_H: str = "&#72;"  #: H
    UPPERCASE_I: str = "&#73;"  #: I
    UPPERCASE_J: str = "&#74;"  #: J
    UPPERCASE_K: str = "&#75;"  #: K
    UPPERCASE_L: str = "&#76;"  #: L
    UPPERCASE_M: str = "&#77;"  #: M
    UPPERCASE_N: str = "&#78;"  #: N
    UPPERCASE_O: str = "&#79;"  #: O
    UPPERCASE_P: str = "&#80;"  #: P
    UPPERCASE_Q: str = "&#81;"  #: Q
    UPPERCASE_R: str = "&#82;"  #: R
    UPPERCASE_S: str = "&#83;"  #: S
    UPPERCASE_T: str = "&#84;"  #: T
    UPPERCASE_U: str = "&#85;"  #: U
    UPPERCASE_V: str = "&#86;"  #: V
    UPPERCASE_W: str = "&#87;"  #: W
    UPPERCASE_X: str = "&#88;"  #: X
    UPPERCASE_Y: str = "&#89;"  #: Y
    UPPERCASE_Z: str = "&#90;"  #: Z

    OPENING_SQUARE_BRACKET: str = "&#91;"  #: [
    BACKSLASH: str = "&#92;"  #: \
    CLOSING_SQUARE_BRACKET: str = "&#93;"  #: ]
    CARET: str = "&#94;"  #: ^
    UNDERSCORE: str = "&#95;"  #: _
    GRAVE_ACCENT: str = "&#96;"  #:
    LOWERCASE_A: str = "&#97;"  #: a
    LOWERCASE_B: str = "&#98;"  #: b
    LOWERCASE_C: str = "&#99;"  #: c
    LOWERCASE_D: str = "&#100;"  #: d
    LOWERCASE_E: str = "&#101;"  #: e
    LOWERCASE_F: str = "&#102;"  #: f
    LOWERCASE_G: str = "&#103;"  #: g
    LOWERCASE_H: str = "&#104;"  #: h
    LOWERCASE_I: str = "&#105;"  #: i
    LOWERCASE_J: str = "&#106;"  #: j
    LOWERCASE_K: str = "&#107;"  #: k
    LOWERCASE_L: str = "&#108;"  #: l
    LOWERCASE_M: str = "&#109;"  #: m
    LOWERCASE_N: str = "&#110;"  #: n
    LOWERCASE_O: str = "&#111;"  #: o
    LOWERCASE_P: str = "&#112;"  #: p
    LOWERCASE_Q: str = "&#113;"  #: q
    LOWERCASE_R: str = "&#114;"  #: r
    LOWERCASE_S: str = "&#115;"  #: s
    LOWERCASE_T: str = "&#116;"  #: t
    LOWERCASE_U: str = "&#117;"  #: u
    LOWERCASE_V: str = "&#118;"  #: v
    LOWERCASE_W: str = "&#119;"  #: w
    LOWERCASE_X: str = "&#120;"  #: x
    LOWERCASE_Y: str = "&#121;"  #: y
    LOWERCASE_Z: str = "&#122;"  #: z

    OPENING_CURLY_BRACE: str = "&#123;"  #: {
    LEFT_CURLY_BRACE: str = "&#123;"  #: {
    VERTICAL_BAR: str = "&#124;"  #: |
    CLOSING_CURLY_BRACE: str = "&#125;"  #: }
    RIGHT_CURLY_BRACE: str = "&#125;"  #: }
    TILDE: str = "&#126;"  #: ~

    # ISO-8859-1 Characters
    AGRAVE: str = "&Agrave;"  #: À
    AACUTE: str = "&Aacute;"  #: Á
    ACIRC: str = "&Acirc;"  #: Â
    ATILDE: str = "&Atilde;"  #: Ã
    AUML: str = "&Auml;"  #: Ä
    ARING: str = "&Aring;"  #: Å
    AELIG: str = "&AElig;"  #: Æ
    CCEDIL: str = "&Ccedil;"  #: Ç
    EGRAVE: str = "&Egrave;"  #: È
    EACUTE: str = "&Eacute;"  #: É
    ECIRC: str = "&Ecirc;"  #: Ê
    EUML: str = "&Euml;"  #: Ë
    IGRAVE: str = "&Igrave;"  #: Ì
    IACUTE: str = "&Iacute;"  #: Í
    ICIRC: str = "&Icirc;"  #: Î
    IUML: str = "&Iuml;"  #: Ï
    ETH: str = "&ETH;"  #: Ð
    NTILDE: str = "&Ntilde;"  #: Ñ
    OGRAVE: str = "&Ograve;"  #: Ò
    OACUTE: str = "&Oacute;"  #: Ó
    OCIRC: str = "&Ocirc;"  #: Ô
    OTILDE: str = "&Otilde;"  #: Õ
    OUML: str = "&Ouml;"  #: Ö
    OSLASH: str = "&Oslash;"  #: Ø
    UGRAVE: str = "&Ugrave;"  #: Ù
    UACUTE: str = "&Uacute;"  #: Ú
    UCIRC: str = "&Ucirc;"  #: Û
    UUML: str = "&Uuml;"  #: Ü
    YACUTE: str = "&Yacute;"  #: Ý
    THORN: str = "&THORN;"  #: Þ
    SZLIG: str = "&szlig;"  #: ß
    AGRAVE: str = "&agrave;"  #: à
    AACUTE: str = "&aacute;"  #: á
    ACIRC: str = "&acirc;"  #: â
    ATILDE: str = "&atilde;"  #: ã
    AUML: str = "&auml;"  #: ä
    ARING: str = "&aring;"  #: å
    AELIG: str = "&aelig;"  #: æ
    CCEDIL: str = "&ccedil;"  #: ç
    EGRAVE: str = "&egrave;"  #: è
    EACUTE: str = "&eacute;"  #: é
    ECIRC: str = "&ecirc;"  #: ê
    EUML: str = "&euml;"  #: ë
    IGRAVE: str = "&igrave;"  #: ì
    IACUTE: str = "&iacute;"  #: í
    ICIRC: str = "&icirc;"  #: î
    IUML: str = "&iuml;"  #: ï
    ETH: str = "&eth;"  #: ð
    NTILDE: str = "&ntilde;"  #: ñ
    OGRAVE: str = "&ograve;"  #: ò
    OACUTE: str = "&oacute;"  #: ó
    OCIRC: str = "&ocirc;"  #: ô
    OTILDE: str = "&otilde;"  #: õ
    OUML: str = "&ouml;"  #: ö
    OSLASH: str = "&oslash;"  #: ø
    UGRAVE: str = "&ugrave;"  #: ù
    UACUTE: str = "&uacute;"  #: ú
    UCIRC: str = "&ucirc;"  #: û
    UUML: str = "&uuml;"  #: ü
    YACUTE: str = "&yacute;"  #: ý
    THORN: str = "&thorn;"  #: þ
    YUML: str = "&yuml;"  #: ÿ

    # ISO-8859-1 Symbols
    NBSP: str = "&nbsp;"  #:
    IEXCL: str = "&iexcl;"  #: ¡
    CENT: str = "&cent;"  #: ¢
    POUND: str = "&pound;"  #: £
    CURREN: str = "&curren;"  #: ¤
    YEN: str = "&yen;"  #: ¥
    BRVBAR: str = "&brvbar;"  #: ¦
    SECT: str = "&sect;"  #: §
    UML: str = "&uml;"  #: ¨
    COPY: str = "&copy;"  #: ©
    COPYRIGHT: str = "&copy;"  #: ©
    ORDF: str = "&ordf;"  #: ª
    LAQUO: str = "&laquo;"  #: «
    NOT: str = "&not;"  #: ¬
    # ­   &shy;   &#173;  Soft hyphen
    REG: str = "&reg;"  #: ®
    MACR: str = "&macr;"  #: ¯
    DEG: str = "&deg;"  #: °
    PLUSMN: str = "&plusmn;"  #: ±
    SUP2: str = "&sup2;"  #: ²
    SUP3: str = "&sup3;"  #: ³
    ACUTE: str = "&acute;"  #: ´
    MICRO: str = "&micro;"  #: µ
    PARA: str = "&para;"  #: ¶
    CEDIL: str = "&cedil;"  #: ¸
    SUP1: str = "&sup1;"  #: ¹
    ORDM: str = "&ordm;"  #: º
    RAQUO: str = "&raquo;"  #: »
    FRAC14: str = "&frac14;"  #: ¼
    FRAC12: str = "&frac12;"  #: ½
    FRAC34: str = "&frac34;"  #: ¾
    IQUEST: str = "&iquest;"  #: ¿
    TIMES: str = "&times;"  #: ×
    DIVIDE: str = "&divide;"  #: ÷

    # Math Symbols
    FORALL: str = "&forall;"  #: ∀
    PART: str = "&part;"  #: ∂
    EXIST: str = "&exist;"  #: ∃
    EMPTY: str = "&empty;"  #: ∅
    NABLA: str = "&nabla;"  #: ∇
    ISIN: str = "&isin;"  #: ∈
    NOTIN: str = "&notin;"  #: ∉
    NI: str = "&ni;"  #: ∋
    PROD: str = "&prod;"  #: ∏
    SUM: str = "&sum;"  #: ∑
    MINUS: str = "&minus;"  #: −
    LOWAST: str = "&lowast;"  #: ∗
    RADIC: str = "&radic;"  #: √
    PROP: str = "&prop;"  #: ∝
    INFIN: str = "&infin;"  #: ∞
    ANG: str = "&ang;"  #: ∠
    AND: str = "&and;"  #: ∧
    OR: str = "&or;"  #: ∨
    CAP: str = "&cap;"  #: ∩
    CUP: str = "&cup;"  #: ∪
    INT: str = "&int;"  #: ∫
    THERE4: str = "&there4;"  #: ∴
    SIM: str = "&sim;"  #: ∼
    CONG: str = "&cong;"  #: ≅
    ASYMP: str = "&asymp;"  #: ≈
    NE: str = "&ne;"  #: ≠
    EQUIV: str = "&equiv;"  #: ≡
    LE: str = "&le;"  #: ≤
    GE: str = "&ge;"  #: ≥
    SUB: str = "&sub;"  #: ⊂
    SUP: str = "&sup;"  #: ⊃
    NSUB: str = "&nsub;"  #: ⊄
    SUBE: str = "&sube;"  #: ⊆
    SUPE: str = "&supe;"  #: ⊇
    OPLUS: str = "&oplus;"  #: ⊕
    OTIMES: str = "&otimes;"  #: ⊗
    PERP: str = "&perp;"  #: ⊥
    SDOT: str = "&sdot;"  #: ⋅

    # Greek Letters
    ALPHA: str = "&Alpha;"  #: Α
    BETA: str = "&Beta;"  #: Β
    GAMMA: str = "&Gamma;"  #: Γ
    DELTA: str = "&Delta;"  #: Δ
    EPSILON: str = "&Epsilon;"  #: Ε
    ZETA: str = "&Zeta;"  #: Ζ
    ETA: str = "&Eta;"  #: Η
    THETA: str = "&Theta;"  #: Θ
    IOTA: str = "&Iota;"  #: Ι
    KAPPA: str = "&Kappa;"  #: Κ
    LAMBDA: str = "&Lambda;"  #: Λ
    MU: str = "&Mu;"  #: Μ
    NU: str = "&Nu;"  #: Ν
    XI: str = "&Xi;"  #: Ξ
    OMICRON: str = "&Omicron;"  #: Ο
    PI: str = "&Pi;"  #: Π
    RHO: str = "&Rho;"  #: Ρ
    SIGMA: str = "&Sigma;"  #: Σ
    TAU: str = "&Tau;"  #: Τ
    UPSILON: str = "&Upsilon;"  #: Υ
    PHI: str = "&Phi;"  #: Φ
    CHI: str = "&Chi;"  #: Χ
    PSI: str = "&Psi;"  #: Ψ
    OMEGA: str = "&Omega;"  #: Ω
    ALPHA: str = "&alpha;"  #: α
    BETA: str = "&beta;"  #: β
    GAMMA: str = "&gamma;"  #: γ
    DELTA: str = "&delta;"  #: δ
    EPSILON: str = "&epsilon;"  #: ε
    ZETA: str = "&zeta;"  #: ζ
    ETA: str = "&eta;"  #: η
    THETA: str = "&theta;"  #: θ
    IOTA: str = "&iota;"  #: ι
    KAPPA: str = "&kappa;"  #: κ
    LAMBDA: str = "&lambda;"  #: λ
    MU: str = "&mu;"  #: μ
    NU: str = "&nu;"  #: ν
    XI: str = "&xi;"  #: ξ
    OMICRON: str = "&omicron;"  #: ο
    PI: str = "&pi;"  #: π
    RHO: str = "&rho;"  #: ρ
    SIGMAF: str = "&sigmaf;"  #: ς
    SIGMA: str = "&sigma;"  #: σ
    TAU: str = "&tau;"  #: τ
    UPSILON: str = "&upsilon;"  #: υ
    PHI: str = "&phi;"  #: φ
    CHI: str = "&chi;"  #: χ
    PSI: str = "&psi;"  #: ψ
    OMEGA: str = "&omega;"  #: ω
    THETASYM: str = "&thetasym;"  #: ϑ
    UPSIH: str = "&upsih;"  #: ϒ
    PIV: str = "&piv;"  #: ϖ

    OELIG: str = "&OElig;"  #: Œ
    oeLIG: str = "&oelig;"  #: œ
    SCARON: str = "&Scaron;"  #: Š
    Scaron: str = "&Scaron;"  #: Š
    scaron: str = "&scaron;"  #: š
    YUML: str = "&Yuml;"  #: Ÿ
    FNOF: str = "&fnof;"  #: ƒ
    CIRC: str = "&circ;"  #: ˆ
    TILDE: str = "&tilde;"  #: ˜

    #     &ensp;  &#8194; En space
    #     &emsp;  &#8195; Em space
    #     &thinsp;    &#8201; Thin space
    # ‌   &zwnj;  &#8204; Zero width non-joiner
    # ‍   &zwj;   &#8205; Zero width joiner
    # ‎   &lrm;   &#8206; Left-to-right mark
    # ‏   &rlm;   &#8207; Right-to-left mark

    NDASH: str = "&ndash;"  #: –
    MDASH: str = "&mdash;"  #: —
    LSQUO: str = "&lsquo;"  #: ‘
    RSQUO: str = "&rsquo;"  #: ’
    SBQUO: str = "&sbquo;"  #: ‚
    LDQUO: str = "&ldquo;"  #: “
    RDQUO: str = "&rdquo;"  #: ”
    BDQUO: str = "&bdquo;"  #: „
    DAGGER: str = "&dagger;"  #: †
    DAGGER: str = "&Dagger;"  #: ‡
    BULL: str = "&bull;"  #: •
    HELLIP: str = "&hellip;"  #: …
    PERMIL: str = "&permil;"  #: ‰
    PRIME: str = "&prime;"  #: ′
    PRIME: str = "&Prime;"  #: ″
    LSAQUO: str = "&lsaquo;"  #: ‹
    RSAQUO: str = "&rsaquo;"  #: ›
    OLINE: str = "&oline;"  #: ‾
    EURO: str = "&euro;"  #: €
    TRADE: str = "&trade;"  #: ™
    TRADEMARK: str = "&trade;"  #: ™

    # ARROWS
    LARR: str = "&larr;"  #: ←
    LEFT: str = "&larr;"  #: ←
    UARR: str = "&uarr;"  #: ↑
    UP: str = "&uarr;"  #: ↑
    RARR: str = "&rarr;"  #: →
    RIGHT: str = "&rarr;"  #: →
    DARR: str = "&darr;"  #: ↓
    DOWN: str = "&darr;"  #: ↓

    HARR: str = "&harr;"  #: ↔
    CRARR: str = "&crarr;"  #: ↵
    LCEIL: str = "&lceil;"  #: ⌈
    RCEIL: str = "&rceil;"  #: ⌉
    LFLOOR: str = "&lfloor;"  #: ⌊
    RFLOOR: str = "&rfloor;"  #: ⌋
    LOZ: str = "&loz;"  #: ◊

    SPADES: str = "&spades;"  #: ♠
    CLUBS: str = "&clubs;"  #: ♣
    HEARTS: str = "&hearts;"  #: ♥
    DIAMS: str = "&diams;"  #: ♦
    DIAMONDS: str = "&diams;"  #: ♦

    SUNG: str = "&sung;"  #: ♪
    FLAT: str = "&flat;"  #: ♭
    NATUR: str = "&natur;"  #: ♮
    NATURAL: str = "&natural;"  #: ♮
    SHARP: str = "&sharp;"  #: ♯

    CHECK: str = "&check;"  #: ✓
    CHECKMARK: str = "&checkmark;"  #: ✓
    TICK: str = "&check;"  #: ✓
    CROSS: str = "&cross;"  #: ✗

    OHM: str = "&ohm;"  #: Ω
    MHO: str = "&mho;"  #: ℧

    FRAC13: str = "&frac13;"  #: ⅓
    FRAC23: str = "&frac23;"  #: ⅔
    FRAC15: str = "&frac15;"  #: ⅕
    FRAC25: str = "&frac25;"  #: ⅖
    FRAC35: str = "&frac35;"  #: ⅗
    FRAC45: str = "&frac45;"  #: ⅘
    FRAC16: str = "&frac16;"  #: ⅙
    FRAC56: str = "&frac56;"  #: ⅚
    FRAC18: str = "&frac18;"  #: ⅛
    FRAC38: str = "&frac38;"  #: ⅜
    FRAC58: str = "&frac58;"  #: ⅝
    FRAC78: str = "&frac78;"  #: ⅞

    STAR: str = "&star;"  #: ☆
    STARF: str = "&starf;"  #: ★
    BIGSTAR: str = "&bigstar;"
    PHONE: str = "&phone;"  #: ☎
    FEMALE: str = "&female;"  #: ♀
    MALE: str = "&male;"  #: ♂


_CHAR_DISAMBIGUATED_ALIASES = {
    "LATIN_CAPITAL_A_GRAVE": "&Agrave;",
    "LATIN_CAPITAL_A_ACUTE": "&Aacute;",
    "LATIN_CAPITAL_A_CIRCUMFLEX": "&Acirc;",
    "LATIN_CAPITAL_A_TILDE": "&Atilde;",
    "LATIN_CAPITAL_A_UMLAUT": "&Auml;",
    "LATIN_CAPITAL_A_RING": "&Aring;",
    "LATIN_CAPITAL_AE": "&AElig;",
    "LATIN_CAPITAL_C_CEDILLA": "&Ccedil;",
    "LATIN_CAPITAL_E_GRAVE": "&Egrave;",
    "LATIN_CAPITAL_E_ACUTE": "&Eacute;",
    "LATIN_CAPITAL_E_CIRCUMFLEX": "&Ecirc;",
    "LATIN_CAPITAL_E_UMLAUT": "&Euml;",
    "LATIN_CAPITAL_I_GRAVE": "&Igrave;",
    "LATIN_CAPITAL_I_ACUTE": "&Iacute;",
    "LATIN_CAPITAL_I_CIRCUMFLEX": "&Icirc;",
    "LATIN_CAPITAL_I_UMLAUT": "&Iuml;",
    "LATIN_CAPITAL_ETH": "&ETH;",
    "LATIN_CAPITAL_N_TILDE": "&Ntilde;",
    "LATIN_CAPITAL_O_GRAVE": "&Ograve;",
    "LATIN_CAPITAL_O_ACUTE": "&Oacute;",
    "LATIN_CAPITAL_O_CIRCUMFLEX": "&Ocirc;",
    "LATIN_CAPITAL_O_TILDE": "&Otilde;",
    "LATIN_CAPITAL_O_UMLAUT": "&Ouml;",
    "LATIN_CAPITAL_O_STROKE": "&Oslash;",
    "LATIN_CAPITAL_U_GRAVE": "&Ugrave;",
    "LATIN_CAPITAL_U_ACUTE": "&Uacute;",
    "LATIN_CAPITAL_U_CIRCUMFLEX": "&Ucirc;",
    "LATIN_CAPITAL_U_UMLAUT": "&Uuml;",
    "LATIN_CAPITAL_Y_ACUTE": "&Yacute;",
    "LATIN_CAPITAL_THORN": "&THORN;",
    "GREEK_CAPITAL_ALPHA": "&Alpha;",
    "GREEK_CAPITAL_BETA": "&Beta;",
    "GREEK_CAPITAL_GAMMA": "&Gamma;",
    "GREEK_CAPITAL_DELTA": "&Delta;",
    "GREEK_CAPITAL_EPSILON": "&Epsilon;",
    "GREEK_CAPITAL_ZETA": "&Zeta;",
    "GREEK_CAPITAL_ETA": "&Eta;",
    "GREEK_CAPITAL_THETA": "&Theta;",
    "GREEK_CAPITAL_IOTA": "&Iota;",
    "GREEK_CAPITAL_KAPPA": "&Kappa;",
    "GREEK_CAPITAL_LAMBDA": "&Lambda;",
    "GREEK_CAPITAL_MU": "&Mu;",
    "GREEK_CAPITAL_NU": "&Nu;",
    "GREEK_CAPITAL_XI": "&Xi;",
    "GREEK_CAPITAL_OMICRON": "&Omicron;",
    "GREEK_CAPITAL_PI": "&Pi;",
    "GREEK_CAPITAL_RHO": "&Rho;",
    "GREEK_CAPITAL_SIGMA": "&Sigma;",
    "GREEK_CAPITAL_TAU": "&Tau;",
    "GREEK_CAPITAL_UPSILON": "&Upsilon;",
    "GREEK_CAPITAL_PHI": "&Phi;",
    "GREEK_CAPITAL_CHI": "&Chi;",
    "GREEK_CAPITAL_PSI": "&Psi;",
    "GREEK_CAPITAL_OMEGA": "&Omega;",
    "SINGLE_DAGGER": "&dagger;",
    "DOUBLE_DAGGER": "&Dagger;",
    "SINGLE_PRIME": "&prime;",
    "DOUBLE_PRIME": "&Prime;",
    "ASCII_TILDE": "&#126;",
    "LATIN_SMALL_Y_UMLAUT": "&yuml;",
    "LATIN_CAPITAL_Y_UMLAUT": "&Yuml;",
}

for _alias_name, _alias_value in _CHAR_DISAMBIGUATED_ALIASES.items():
    setattr(Char, _alias_name, _alias_value)


__all__ = ["Char", "Entity"]



"""
    domonic.ext
    ====================================

    This dir is for extending domonic to be useable with other python libs.

"""

from __future__ import annotations

# HELLO WORLDS - hello world code for other libs
# "django"
# "bottle"
# "pyramid"
# "werkzeug"
# "tornado"
# "aiohttp"
# "fastapi"
# "starlette"
# "blacksheep"
# "muffin"
# "falcon"
# "baize"
# "emmett"
# "quart"

# "falcon"
# "eve"
# "graphene"
# "httpx"
# "invenio"
# "jupyterhub"
# "klein"
# "kombu"
# "masonite"
# "motor"
# "pydantic"
# "quart"
# "sanic"
# "trio"


HELLO_BLACKSHEEP: str = """
import uvicorn
from domonic.html import *

from blacksheep.server import Application
from blacksheep.server.responses import html as HTMLResponse

app = Application()

@app.route("/")
async def home(request):
    return HTMLResponse(str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    ))

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)
"""

HELLO_FAST_API: str = """
import uvicorn
from fastapi import FastAPI
from fastapi.responses import HTMLResponse
from domonic.html import *

app = FastAPI()

@app.get("/")
def read_root():
    return HTMLResponse(str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    ))

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)
"""

HELLO_STARLETTE: str = """
import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route
from domonic.html import *

async def homepage(request):
    return HTMLResponse(str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    ))

app = Starlette(routes=[Route("/", homepage)])

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)
"""

HELLO_SANIC: str = """
from sanic import Sanic
from sanic import response
from domonic.html import *

app = Sanic("My Hello, world app")

@app.route('/')
async def test(request):
    return response.html(str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        ))
    )

if __name__ == '__main__':
    app.run()
"""

HELLO_FLASK: str = """
from flask import Flask
from domonic.html import *

app = Flask(__name__)

@app.route("/")
def hello():
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        ))

if __name__ == '__main__':
    app.run(host="127.0.0.1", port=5000)
"""

HELLO_CHERRYPY: str = """
import cherrypy
from domonic.html import *

class HelloWorld:

    @cherrypy.expose
    def index(self):
        return str(
                    html(
                    head(),
                    body(
                        div(span("Hello, World!"))
                        )
                    )
                )

cherrypy.quickstart(HelloWorld())
"""


HELLO_BOTTLE: str = """
from bottle import route, run
from domonic.html import *

@route('/hello/<name>')
def index(name):
    return str(
            html(
            head(),
            body(
                div(span(f"Hello, {name}!"))
                )
            )
        )

run(host='localhost', port=8080)
"""

HELLO_DJANGO: str = """
import django
from django.conf import settings
from django.core.management import execute_from_command_line
from django.http import HttpResponse
from django.urls import path
from domonic.html import *

def index(request):
    page = html(head(), body(div(span("Hello World!"))))
    return HttpResponse(str(page))

urlpatterns = [path("", index)]

if not settings.configured:
    settings.configure(
        DEBUG=True,
        ROOT_URLCONF=__name__,
        SECRET_KEY="domonic-cli",
        ALLOWED_HOSTS=["*"],
        MIDDLEWARE=[],
        INSTALLED_APPS=[],
    )
    django.setup()

if __name__ == "__main__":
    execute_from_command_line(["app.py", "runserver", "127.0.0.1:8000"])
"""

HELLO_PYRAMID: str = """
from wsgiref.simple_server import make_server
from pyramid.config import Configurator
from pyramid.response import Response
from domonic.html import *

def home(request):
    page = html(head(), body(div(span("Hello World!"))))
    return Response(str(page), content_type="text/html")

if __name__ == "__main__":
    with Configurator() as config:
        config.add_route("home", "/")
        config.add_view(home, route_name="home")
        app = config.make_wsgi_app()
    server = make_server("127.0.0.1", 6543, app)
    server.serve_forever()
"""

HELLO_AIOHTTP: str = """
from domonic.html import *
from aiohttp import web

async def handle(request):
    name = request.match_info.get('name', "Anonymous")
    page = html(head(),body(div(span("Hello, World!"))))
    return web.Response(text=str(page), content_type='text/html')

app = web.Application()
app.add_routes([web.get('/', handle),
                web.get('/{name}', handle)])

if __name__ == '__main__':
    web.run_app(app)
"""

HELLO_TORNADO: str = """
import tornado.ioloop
import tornado.web
from domonic.html import *

class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write(str(
            html(
            head(),
            body(
                div(span("Hello World!"))
                )
            )))

def make_app():
    return tornado.web.Application([
        (r"/", MainHandler),
    ])

if __name__ == "__main__":
    app = make_app()
    app.listen(8888)
    tornado.ioloop.IOLoop.current().start()
"""

HELLO_WERKZEUG: str = """
from werkzeug.wrappers import Request, Response
from domonic.html import *

@Request.application
def application(request):
    return Response(str(
                    html(
                    head(),
                    body(
                        div(span("Hello World!"))
                        )
                    )), mimetype='text/html')

if __name__ == '__main__':
    from werkzeug.serving import run_simple
    run_simple('localhost', 4000, application)
"""

HELLO_FALCON: str = """
import falcon
from wsgiref.simple_server import make_server
from domonic.html import *

class HomeResource:

    def on_get(self, req, resp):
        page = html(head(), body(div(span("Hello World!"))))
        resp.content_type = "text/html"
        resp.text = str(page)

app = falcon.App()
app.add_route("/", HomeResource())

if __name__ == "__main__":
    server = make_server("127.0.0.1", 8000, app)
    server.serve_forever()
"""

HELLO_QUART: str = """
from quart import Quart
from domonic.html import *

app = Quart(__name__)

@app.route("/")
async def hello():
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    )

if __name__ == "__main__":
    app.run(host="127.0.0.1", port=8000)
"""

HELLO_MUFFIN: str = """
import muffin
import uvicorn
from domonic.html import *

app = muffin.Application()

@app.route("/")
async def hello(request):
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    )

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)
"""

HELLO_BAIZE: str = """
import uvicorn
from baize.asgi import HTMLResponse, request_response
from domonic.html import *

@request_response
async def app(request):
    page = html(head(), body(div(span("Hello World!"))))
    return HTMLResponse(str(page))

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)
"""

HELLO_EMMETT: str = """
from emmett import App
from domonic.html import *

app = App(__name__)

@app.route("/")
async def index():
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    )

if __name__ == "__main__":
    app.run(host="127.0.0.1", port=8000)
"""

HELLO_LITESTAR: str = """
import uvicorn
from litestar import Litestar, get
from domonic.html import *

@get("/")
async def index() -> str:
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    )

app = Litestar(route_handlers=[index])

if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8000)
"""

HELLO_ROBYN: str = """
from robyn import Robyn
from domonic.html import *

app = Robyn(__file__)

@app.get("/")
async def index(request):
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        )
    )

app.start(port=8080)
"""


SERVER_SCAFFOLDS: dict[str, dict[str, object]] = {
    "none": {"packages": [], "template": None},
    "sanic": {"packages": ["sanic==25.12.0"], "template": HELLO_SANIC},
    "flask": {"packages": ["Flask==3.1.3"], "template": HELLO_FLASK},
    "cherrypy": {"packages": ["CherryPy==18.10.0"], "template": HELLO_CHERRYPY},
    "django": {
        "packages": [
            'Django==5.2.12; python_version < "3.12"',
            'Django==6.0.3; python_version >= "3.12"',
        ],
        "template": HELLO_DJANGO,
    },
    "bottle": {"packages": ["bottle==0.13.4"], "template": HELLO_BOTTLE},
    "pyramid": {"packages": ["pyramid==2.1"], "template": HELLO_PYRAMID},
    "werkzeug": {"packages": ["Werkzeug==3.1.7"], "template": HELLO_WERKZEUG},
    "tornado": {"packages": ["tornado==6.5.5"], "template": HELLO_TORNADO},
    "aiohttp": {"packages": ["aiohttp==3.13.3"], "template": HELLO_AIOHTTP},
    "fastapi": {"packages": ["fastapi==0.135.2", "uvicorn==0.42.0"], "template": HELLO_FAST_API},
    "starlette": {"packages": ["starlette==1.0.0", "uvicorn==0.42.0"], "template": HELLO_STARLETTE},
    "blacksheep": {"packages": ["blacksheep==2.6.2", "uvicorn==0.42.0"], "template": HELLO_BLACKSHEEP},
    "muffin": {"packages": ["muffin==1.8.2", "uvicorn==0.42.0"], "template": HELLO_MUFFIN},
    "falcon": {"packages": ["falcon==4.2.0"], "template": HELLO_FALCON},
    "baize": {"packages": ["baize==0.23.1", "uvicorn==0.42.0"], "template": HELLO_BAIZE},
    "emmett": {"packages": ["emmett==2.7.1"], "template": HELLO_EMMETT},
    "litestar": {"packages": ["litestar==2.21.1", "uvicorn==0.42.0"], "template": HELLO_LITESTAR},
    "quart": {"packages": ["Quart==0.20.0", "Hypercorn==0.18.0"], "template": HELLO_QUART},
    "robyn": {"packages": ["robyn==0.82.1"], "template": HELLO_ROBYN},
}


def get_supported_servers() -> list[str]:
    return list(SERVER_SCAFFOLDS.keys())


def get_server_requirements(server: str) -> list[str]:
    scaffold = SERVER_SCAFFOLDS.get(server, {})
    return list(scaffold.get("packages", []))


def get_hello_world(server: str) -> str | None:
    scaffold = SERVER_SCAFFOLDS.get(server, {})
    return scaffold.get("template")



"""
    test_events
    ~~~~~~~~~~~~
    - unit tests for events

"""

import asyncio
import unittest

from domonic import *
from domonic.dom import Range, document
from domonic.events import *

class TestCase(unittest.TestCase):
    """Tests for the events classes"""

    def test_html_structure(self):
        site = html()
        somebody = document.createElement("div")
        site.appendChild(somebody)
        self.assertEqual(str(site), "<html><div></div></html>")

    def test_event_handling(self):
        site = html()
        somebody = document.createElement("div")
        site.appendChild(somebody)

        event_results = []

        def test(evt, *args, **kwargs):
            event_results.append(evt.target)

        site.addEventListener("click", test)
        somebody.addEventListener("anything", test)

        self.assertIn("click", site.listeners)
        self.assertEqual(site.listeners["click"], [test])

        site.dispatchEvent(Event("click"))
        somebody.dispatchEvent(Event("anything"))

        self.assertIn(somebody, event_results)
        self.assertIn(site, event_results)

    def test_events(self):
        site = EventTarget()
        somebody = EventTarget()

        def test(evt, *args, **kwargs):
            self.assertTrue(evt.target == somebody or evt.target == site)

        async def async_test(evt, *args, **kwargs):
            await asyncio.sleep(0)
            self.assertTrue(evt.target == somebody or evt.target == site)

        site.addEventListener("click", test)
        somebody.addEventListener("anything", test)

        site.dispatchEvent(Event("click"))
        somebody.dispatchEvent(Event("anything"))

        # Asynchronous event handling
        site.addEventListener("async_event", async_test)
        asyncio.run(site.dispatchEventAsync(Event("async_event")))

    def test_async_events(self):
        async def runner():
            site = EventTarget()
            somebody = EventTarget()

            async def async_test(evt, *args, **kwargs):
                await asyncio.sleep(0)
                self.assertTrue(evt.target == somebody or evt.target == site)

            site.addEventListener("click", async_test)
            somebody.addEventListener("anything", async_test)

            await site.dispatchEventAsync(Event("click"))
            await somebody.dispatchEventAsync(Event("anything"))

        asyncio.run(runner())

    def test_default_event_properties(self):
        event = Event("custom_event")
        self.assertIsNone(event.target)
        self.assertFalse(event.defaultPrevented)
        self.assertEqual(event.eventPhase, Event.NONE)
        self.assertFalse(event.bubbles)
        self.assertFalse(event.cancelable)

    def test_custom_event_properties(self):
        # Define custom event data
        event_data = {"message": "Hello, world!"}

        # Create a custom event
        event = Event("custom_event")
        event.custom_data = event_data

        # Test custom event properties
        self.assertEqual(event.type, "custom_event")
        self.assertFalse(event.bubbles)
        self.assertFalse(event.cancelable)
        self.assertEqual(event.eventPhase, Event.NONE)
        self.assertIsInstance(event.timeStamp, (int, float))
        self.assertEqual(event.custom_data, event_data)

    def test_prevent_default(self):
        event = Event("custom_event", {"cancelable": True})
        event.preventDefault()
        self.assertTrue(event.defaultPrevented)

    def test_stop_propagation(self):
        event = Event("custom_event")
        event.stopImmediatePropagation()
        self.assertTrue(event.cancelBubble)

    def test_prevent_default_respects_cancelable(self):
        event = Event("custom_event", {"cancelable": False})
        event.preventDefault()
        self.assertFalse(event.defaultPrevented)

    def test_event_bubbling(self):
        # Create an event target and some event handlers
        target = EventTarget()
        results = []

        def event_handler1(event):
            results.append("Handler 1")

        def event_handler2(event):
            results.append("Handler 2")

        def event_handler3(event):
            results.append("Handler 3")

        # Add event listeners for a custom event
        target.addEventListener("custom_event", event_handler1)
        target.addEventListener("custom_event", event_handler2)
        target.addEventListener("custom_event", event_handler3)

        # Create a custom event
        event_data = {"message": "Hello, world!"}
        custom_event = Event("custom_event", data=event_data)

        # Dispatch the event
        target.dispatchEvent(custom_event)

        # Check if event bubbling occurred in the correct order
        expected_results = ["Handler 1", "Handler 2", "Handler 3"]
        self.assertEqual(results, expected_results)

    def test_event_bubbles_to_parent_targets(self):
        parent = EventTarget()
        child = EventTarget()
        child.parentNode = parent
        results = []

        parent.addEventListener("custom_event", lambda event: results.append(("parent", event.eventPhase)))
        child.addEventListener("custom_event", lambda event: results.append(("child", event.eventPhase)))

        child.dispatchEvent(Event("custom_event", {"bubbles": True}))

        self.assertEqual(results, [("child", Event.AT_TARGET), ("parent", Event.BUBBLING_PHASE)])

    def test_event_target_with_target_matching(self):
        # Create an event target and some event handlers
        target = EventTarget()
        results = []

        def event_handler1(event):
            results.append("Handler 1")

        def event_handler2(event):
            results.append("Handler 2")

        def event_handler3(event):
            results.append("Handler 3")

        def specific_target_handler(event):
            results.append("Specific Target Handler")

        # Add event listeners for custom events with different targets
        target.addEventListener("custom_event", event_handler1)
        target.addEventListener("custom_event", event_handler2)
        target.addEventListener("custom_event", event_handler3)

        specific_target = EventTarget()
        specific_target.addEventListener("custom_event", specific_target_handler)

        # Create a custom event with a specific target
        event_data = {"message": "Hello, world!"}
        custom_event = Event("custom_event", data=event_data)

        # Dispatch the event on the specific target
        specific_target.dispatchEvent(custom_event)

        # Check if the event only triggered the specific target handler
        expected_results = ["Specific Target Handler"]
        self.assertEqual(results, expected_results)

    def test_event_capture_phase(self):
        # Create an event target and some event handlers
        target = EventTarget()
        results = []

        def capture_handler1(event):
            results.append("Capture Handler 1")

        def capture_handler2(event):
            results.append("Capture Handler 2")

        def capture_handler3(event):
            results.append("Capture Handler 3")

        def bubble_handler(event):
            results.append("Bubble Handler")

        # Add event listeners for custom events with the capture phase
        target.addEventListener("custom_event", capture_handler1, use_capture=True)
        target.addEventListener("custom_event", capture_handler2, use_capture=True)
        target.addEventListener("custom_event", capture_handler3, use_capture=True)
        target.addEventListener("custom_event", bubble_handler)

        # Create a custom event
        event_data = {"message": "Hello, world!"}
        custom_event = Event("custom_event", data=event_data)

        # Dispatch the event
        target.dispatchEvent(custom_event)

        # Check if the event listeners in the capture phase were executed in order
        expected_results = ["Capture Handler 1", "Capture Handler 2", "Capture Handler 3", "Bubble Handler"]
        self.assertEqual(results, expected_results)

    def test_capture_and_bubble_order_across_path(self):
        root = EventTarget()
        parent = EventTarget()
        child = EventTarget()
        parent.parentNode = root
        child.parentNode = parent
        results = []

        root.addEventListener("custom_event", lambda event: results.append("root-capture"), {"capture": True})
        parent.addEventListener("custom_event", lambda event: results.append("parent-capture"), {"capture": True})
        child.addEventListener("custom_event", lambda event: results.append("child-capture"), {"capture": True})
        child.addEventListener("custom_event", lambda event: results.append("child-bubble"))
        parent.addEventListener("custom_event", lambda event: results.append("parent-bubble"))

        child.dispatchEvent(Event("custom_event", {"bubbles": True}))

        self.assertEqual(
            results,
            ["root-capture", "parent-capture", "child-capture", "child-bubble", "parent-bubble"],
        )

    def test_once_listener_is_removed_after_first_dispatch(self):
        target = EventTarget()
        calls = []

        target.addEventListener("custom_event", lambda event: calls.append(event.type), {"once": True})

        target.dispatchEvent(Event("custom_event"))
        target.dispatchEvent(Event("custom_event"))

        self.assertEqual(calls, ["custom_event"])

    def test_stop_immediate_propagation_stops_remaining_listeners(self):
        target = EventTarget()
        calls = []

        def first(event):
            calls.append("first")
            event.stopImmediatePropagation()

        def second(event):
            calls.append("second")

        target.addEventListener("custom_event", first)
        target.addEventListener("custom_event", second)

        target.dispatchEvent(Event("custom_event"))

        self.assertEqual(calls, ["first"])

    def test_ui_event_initialization(self):
        event_data = {
            "canBubble": True,
            "cancelable": True,
            "view": "main_window",
            "detail": "click_detail",
            "layerX": 10,
            "layerY": 20,
            "sourceCapabilities": "capabilities"
        }

        ui_event = UIEvent(Event.RESIZE, **event_data)

        self.assertEqual(ui_event.type, "resize")
        self.assertEqual(ui_event.canBubble, True)
        self.assertEqual(ui_event.cancelable, True)
        self.assertEqual(ui_event.view, "main_window")
        self.assertEqual(ui_event.detail, "click_detail")
        self.assertEqual(ui_event.layerX, 10)
        self.assertEqual(ui_event.layerY, 20)
        self.assertEqual(ui_event.sourceCapabilities, "capabilities")

    def test_composed_path(self):
        # Create an event target hierarchy
        root = EventTarget()
        parent = EventTarget()
        child = EventTarget()

        # Simulate the hierarchy
        child.parentNode = parent
        parent.parentNode = root

        # Create a UIEvent and set the target
        event_data = {"type": "click"}
        ui_event = UIEvent("click", options=event_data)
        ui_event.target = child

        # Get the composed path
        path = ui_event.composedPath()

        # Check if the path includes the expected elements in the correct order
        expected_path = [child, parent, root]
        self.assertEqual(path, expected_path)

    def test_custom_event_detail(self):
        event = CustomEvent("custom_event", {"detail": {"message": "hello"}})
        self.assertEqual(event.detail, {"message": "hello"})

    def test_submit_event_submitter(self):
        submitter = object()
        event = SubmitEvent("submit", {"submitter": submitter})
        self.assertIs(event.submitter, submitter)

    def test_mouse_event_client_coordinates_and_modifiers(self):
        event = MouseEvent("click", {"clientX": 10, "clientY": 20, "ctrlKey": True})
        self.assertEqual(event.clientX, 10)
        self.assertEqual(event.clientY, 20)
        self.assertTrue(event.getModifierState("Control"))

    def test_document_create_event_specializations(self):
        self.assertIsInstance(document.createEvent("MouseEvent"), MouseEvent)
        self.assertIsInstance(document.createEvent("KeyboardEvent"), KeyboardEvent)
        self.assertIsInstance(document.createEvent("CustomEvent"), CustomEvent)
        self.assertIsInstance(document.createEvent("PointerEvent"), PointerEvent)
        self.assertIsInstance(document.createEvent("WheelEvent"), WheelEvent)
        self.assertIsInstance(document.createEvent("MessageEvent"), MessageEvent)
        self.assertIsInstance(document.createEvent("ErrorEvent"), ErrorEvent)
        self.assertIsInstance(document.createEvent("PopStateEvent"), PopStateEvent)
        self.assertIsInstance(document.createEvent("CloseEvent"), CloseEvent)

    def test_handle_event_listener_object(self):
        target = EventTarget()
        calls = []

        class Handler:
            def handleEvent(self, event):
                calls.append(event.type)

        target.addEventListener("custom_event", Handler())
        target.dispatchEvent(Event("custom_event"))

        self.assertEqual(calls, ["custom_event"])

    def test_passive_listener_cannot_prevent_default(self):
        target = EventTarget()

        def listener(event):
            event.preventDefault()

        target.addEventListener("custom_event", listener, {"passive": True})
        event = Event("custom_event")
        target.dispatchEvent(event)

        self.assertFalse(event.defaultPrevented)

    def test_abort_controller_signal_and_listener_options(self):
        controller = AbortController()
        signal = controller.signal
        events = []
        target = EventTarget()
        listener_calls = []

        signal.addEventListener("abort", lambda event: events.append((event.type, signal.reason)))
        target.addEventListener("custom_event", lambda event: listener_calls.append(event.type), {"signal": signal})

        controller.abort("stop")
        target.dispatchEvent(Event("custom_event"))

        self.assertTrue(signal.aborted)
        self.assertEqual(signal.reason, "stop")
        self.assertEqual(events, [("abort", "stop")])
        self.assertEqual(listener_calls, [])

        already_aborted = AbortController()
        already_aborted.abort("done")
        target.addEventListener("never", lambda event: listener_calls.append("never"), {"signal": already_aborted.signal})
        self.assertFalse(target.hasEventListener("never"))

    def test_close_event_shape(self):
        event = CloseEvent("close", {"code": 1000, "reason": "bye", "wasClean": True})
        self.assertEqual(event.code, 1000)
        self.assertEqual(event.reason, "bye")
        self.assertTrue(event.wasClean)

    def test_long_tail_event_data_helpers(self):
        custom_event = CustomEvent("custom")
        custom_event.initCustomEvent("customized", bubbles=False, cancelable=True, detail={"ok": True})
        self.assertEqual(custom_event.type, "customized")
        self.assertEqual(custom_event.detail, {"ok": True})
        self.assertFalse(custom_event.bubbles)

        transition = TransitionEvent("transitionend", {"propertyName": "opacity", "elapsedTime": 0.25})
        self.assertEqual(transition.propertyName, "opacity")
        self.assertEqual(transition.elapsedTime, 0.25)

        request = type("Request", (), {"url": "/a", "referrer": "/a", "clientId": "c1"})()
        fetch_event = FetchEvent("fetch", {"request": request, "clientId": "c2"})
        self.assertTrue(fetch_event.isReload)
        self.assertTrue(fetch_event.replacesClientId)
        self.assertEqual(fetch_event.resultingClientId, "c2")
        self.assertEqual(fetch_event.respondWith("response"), "response")

        marker = object()
        extendable = ExtendableEvent("extendable")
        self.assertIs(extendable.waitUntil(marker), marker)
        self.assertEqual(extendable._pending_promises, [marker])

        promise_event = PromiseRejectionEvent("unhandledrejection", {"reason": "boom", "isRejected": True})
        self.assertEqual(promise_event.reason, "boom")
        self.assertTrue(promise_event.isRejected)

        message = MessageEvent("message", {"data": "hi", "ports": [1, 2]})
        self.assertEqual(message.data, "hi")
        self.assertEqual(message.ports, [1, 2])

    def test_default_global_and_window_event_handlers_do_not_throw(self):
        global_handler = GlobalEventHandler()
        click_event = Event("click")
        self.assertIs(global_handler.onclick(click_event), click_event)
        self.assertEqual(global_handler.onkeydown(Event("keydown")).type, "keydown")

        win = WindowEventHandler(window=None)
        resize_event = Event("resize")
        self.assertIs(win.onresize(resize_event), resize_event)
        self.assertEqual(win._last_event.type, "resize")

    def test_event_init_helpers_and_subclass_shape(self):
        ui_event = UIEvent("resize").initUIEvent("scroll", True, False, "view", 2)
        self.assertEqual(ui_event.type, "scroll")
        self.assertTrue(ui_event.bubbles)
        self.assertFalse(ui_event.cancelable)

        mouse_event = MouseEvent("click").initMouseEvent("mousedown", True, True, None, 1, 0, 0, 11, 22, True, False, False, False, 1, None)
        self.assertEqual(mouse_event.type, "mousedown")
        self.assertEqual(mouse_event.clientX, 11)
        self.assertTrue(mouse_event.ctrlKey)

        keyboard_event = KeyboardEvent("keydown").initKeyboardEvent("keyup", True, True, None, 65, "A", 0, "", False)
        self.assertEqual(keyboard_event.type, "keyup")
        self.assertEqual(keyboard_event.key, "a")
        self.assertEqual(keyboard_event.charCode, 65)
        self.assertEqual(keyboard_event.code, "KeyA")
        self.assertEqual(keyboard_event.keyCode, 65)

        keyboard_with_modifiers = KeyboardEvent("keydown").initKeyboardEvent(
            "keydown", True, True, None, 65, "A", 1, "Alt Shift", True
        )
        self.assertTrue(keyboard_with_modifiers.altKey)
        self.assertTrue(keyboard_with_modifiers.shiftKey)
        self.assertEqual(keyboard_with_modifiers.location, 1)
        self.assertTrue(keyboard_with_modifiers.repeat)
        self.assertEqual(keyboard_with_modifiers.code, "KeyA")

    def test_keyboard_event_modern_key_code_and_location_defaults(self):
        keyboard_event = KeyboardEvent(
            "keydown",
            {
                "key": "Enter",
                "location": KeyboardEvent.DOM_KEY_LOCATION_NUMPAD,
                "capsLock": True,
                "repeat": True,
            },
        )

        self.assertEqual(keyboard_event.key, "Enter")
        self.assertEqual(keyboard_event.code, "NumpadEnter")
        self.assertEqual(keyboard_event.keyCode, 13)
        self.assertEqual(keyboard_event.location, KeyboardEvent.DOM_KEY_LOCATION_NUMPAD)
        self.assertTrue(keyboard_event.repeat)
        self.assertTrue(keyboard_event.getModifierState("CapsLock"))
        self.assertFalse(keyboard_event.getModifierState("Shift"))

        pointer_event = PointerEvent("pointerdown", {"clientX": 5, "clientY": 6, "pointerId": 3})
        self.assertEqual(pointer_event.clientX, 5)
        self.assertEqual(pointer_event.pointerId, 3)

        error_event = ErrorEvent("error", {"message": "boom", "filename": "x.py", "lineno": 4, "colno": 2})
        self.assertEqual(error_event.filename, "x.py")
        self.assertEqual(error_event.lineno, 4)

    def test_dispatch_does_not_swallow_listener_exceptions(self):
        target = EventTarget()

        def boom(event):
            raise RuntimeError("boom")

        target.addEventListener("custom_event", boom)
        with self.assertRaises(RuntimeError):
            target.dispatchEvent(Event("custom_event"))

    def test_cancel_bubble_and_return_value_semantics(self):
        event = Event("custom")
        event.cancelBubble = True
        self.assertTrue(event.cancelBubble)
        self.assertTrue(event._propagation_stopped)

        event = Event("custom")
        event.returnValue = False
        self.assertTrue(event.defaultPrevented)
        self.assertFalse(event.returnValue)

        event.returnValue = True
        self.assertFalse(event.defaultPrevented)
        self.assertTrue(event.returnValue)

    def test_pointer_beforeunload_input_and_fetch_helpers(self):
        pointer = PointerEvent("pointermove", {"coalescedEvents": [1, 2], "predictedEvents": [3]})
        self.assertEqual(pointer.getCoalescedEvents(), [1, 2])
        self.assertEqual(pointer.getPredictedEvents(), [3])

        beforeunload = BeforeUnloadEvent("beforeunload")
        beforeunload.returnValue = "Leave?"
        self.assertTrue(beforeunload.defaultPrevented)
        self.assertEqual(beforeunload.returnValue, "Leave?")
        beforeunload.returnValue = ""
        self.assertFalse(beforeunload.defaultPrevented)

        target = EventTarget()
        input_event = InputEvent("input")
        input_event.target = target
        input_event._targetRanges = [Range()]
        self.assertEqual(len(input_event.getTargetRanges()), 1)

        request = type("Request", (), {"url": "/a", "referrer": "/b", "clientId": "c1"})()
        fetch_event = FetchEvent("fetch", {"request": request})
        marker = object()
        self.assertIs(fetch_event.waitUntil(marker), marker)
        self.assertEqual(fetch_event._pending_promises, [marker])

    def test_dispatch_resets_phase_and_current_target(self):
        target = EventTarget()
        event = Event("custom_event", {"bubbles": True, "cancelable": True})

        target.dispatchEvent(event)

        self.assertEqual(event.eventPhase, Event.NONE)
        self.assertIsNone(event.currentTarget)

    def test_async_dispatch_respects_false_and_resets_state(self):
        async def runner():
            target = EventTarget()

            async def listener(event):
                await asyncio.sleep(0)
                return False

            target.addEventListener("custom_event", listener)
            event = Event("custom_event", {"cancelable": True})
            result = await target.dispatchEventAsync(event)
            self.assertFalse(result)
            self.assertTrue(event.defaultPrevented)
            self.assertEqual(event.eventPhase, Event.NONE)
            self.assertIsNone(event.currentTarget)

        asyncio.run(runner())

if __name__ == '__main__':
    unittest.main()



"""
    domonic.webapi.netinfo
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Network_Information_API
"""

from domonic.dom import EventTarget


class NetworkInformation(EventTarget):
    def __init__(self):
        self.type = None
        self.downlinkMax = None
        self.effectiveType = None
        self.rtt = None
        self.saveData = None
        self.downlink = None
        self.effectiveType = None
        self.rtt = None

    def __str__(self):
        return "type: {}, downlinkMax: {}, effectiveType: {}, rtt: {}, saveData: {}, downlink: {}, effectiveType: {}, rtt: {}".format(
            self.type,
            self.downlinkMax,
            self.effectiveType,
            self.rtt,
            self.saveData,
            self.downlink,
            self.effectiveType,
            self.rtt,
        )
